import os
import sys
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np

# sys.path.append('./BIC_codes/')
from functions.dFC_funcs import *
from functions.post_analysis_funcs import *
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

print(
    "################################# POST ANALYSIS STARTED RUNNING ... #################################"
)

################################# LOAD RESULTS #################################

# local code
# assessment_results_root = './../../../../RESULTs/methods_implementation/server/methods_implementation/'
# output_root = './../../../../RESULTs/methods_implementation/server/methods_implementation/out/'

# server code
assessment_results_root = "./"
output_root = "./output/"

FOLDER_name = "similarity_measured/"
num_randomization = 10
num_subj2include = None  # None -> all

ALL_RECORDS = os.listdir(assessment_results_root + FOLDER_name)
ALL_RECORDS = [i for i in ALL_RECORDS if "SUBJ_" in i]
ALL_RECORDS.sort()
for s in ALL_RECORDS[:1]:
    output = np.load(
        assessment_results_root + FOLDER_name + s, allow_pickle="True"
    ).item()

print(f"*** {len(ALL_RECORDS)} subjects were found.")

if not num_subj2include is None:
    ALL_RECORDS = ALL_RECORDS[:num_subj2include]
    print(f"*** {num_subj2include} subjects were included.")

FILTERS = [key for key in output]
print(FILTERS)

for filter in FILTERS:
    measure_name_lst = None
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        # check measures order
        if measure_name_lst is None:
            measure_name_lst = [
                measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
            ]
        else:
            assert measure_name_lst == [
                measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
            ], "measures order mismatch"

# the dictionary that collects all RESULTS
ALL_RESULTS = {}

################################ common variables #################################

for filter in ["default_values"]:

    # for SUBJs_output in SUBJs_output_lst:
    for s in ALL_RECORDS[:1]:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        node_networks = node_info2network(
            SUBJs_output[filter]["TS_info_lst"][0]["nodes_info"]
        )

ALL_RESULTS["node_networks"] = node_networks
ALL_RESULTS["num_subj"] = len(ALL_RECORDS)


################################# dFC SAMPLES #################################
"""
only the first 10 time points are visualized
the rank_norm is performed globally (all time points of each subject together)
"""
for filter in ["default_values"]:

    RESULTS = {}

    # for SUBJs_output in SUBJs_output_lst:
    for s in ALL_RECORDS[:1]:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        for measure_id in SUBJs_output[filter]["dFCM_samples"]:
            TRs = SUBJs_output[filter]["common_TRs"][4:10]

            # rank the whole dFC
            dFC_mat = SUBJs_output[filter]["dFCM_samples"][measure_id]
            dFC_mat_ranked = rank_norm(dFC_mat, global_norm=True)

            samples = {}
            samples_ranked = {}
            for tr in TRs:
                samples["TR" + str(tr)] = dFC_mat[
                    SUBJs_output[filter]["common_TRs"].index(tr), :, :
                ]
                samples_ranked["TR" + str(tr)] = dFC_mat_ranked[
                    SUBJs_output[filter]["common_TRs"].index(tr), :, :
                ]

            measure_name = SUBJs_output[filter]["measure_lst"][
                int(measure_id)
            ].measure_name

            if not measure_name in RESULTS:
                RESULTS[measure_name] = {}

            RESULTS[measure_name]["samples"] = samples
            RESULTS[measure_name]["samples_ranked"] = samples_ranked

ALL_RESULTS["dFC_sample"] = deepcopy(RESULTS)

################################# FCS visualization #################################

for filter in ["default_values"]:

    # for SUBJs_output in SUBJs_output_lst:
    for s in ALL_RECORDS[:1]:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

ALL_RESULTS["measure_lst"] = SUBJs_output[filter]["measure_lst"]

################################# dFC values distributions #################################

RESULTS = {}
for filter in ["default_values"]:
    for s in ALL_RECORDS[:100]:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

            dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]

            if not measure_i.measure_name in RESULTS:
                RESULTS[measure_i.measure_name] = list()
            RESULTS[measure_i.measure_name].append(dFC_mat2vec(dFC_mat_i).flatten())

    for measure in RESULTS:
        RESULTS[measure] = np.array(RESULTS[measure]).flatten()

ALL_RESULTS["dFC_values_dist"] = deepcopy(RESULTS)

################################# dFC Similarity #################################

################# whole subject #################
"""
    - corr((all dFC timepoints of one Subj using method_i), (all dFC timepoints of one Subj using method_j)) -> avg(corr) and var(corr)
    - spearman correlation, pearson correlation, Mutual Information (MI), Euclidean Distance
    - dendogram based on avg(corr)
"""
ALL_RESULTS["dFC_similarity_overall"] = {}

metric_list = ["corr", "spearman", "MI", "euclidean_distance"]
for metric in metric_list:

    RESULTS = {}
    for filter in FILTERS:

        all_subjs_sim_mat = list()
        for s in ALL_RECORDS:
            SUBJs_output = np.load(
                assessment_results_root + FOLDER_name + s, allow_pickle="True"
            ).item()
            # SUBJs_output[filter]['all']['metric'] = (1, method, method)
            all_subjs_sim_mat.append(np.squeeze(SUBJs_output[filter]["all"][metric]))

        measure_name_lst = [
            measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
        ]
        sim_distribution = make_sim_distribution(
            sim_mats_lst=all_subjs_sim_mat, name_lst=measure_name_lst, zip_names=True
        )
        all_subjs_sim_mat = np.array(all_subjs_sim_mat)
        all_subjs_avg = np.mean(all_subjs_sim_mat, axis=0)
        across_subj_var = np.var(all_subjs_sim_mat, axis=0)

        # E(VAR(sim)) across subj
        var_lst = []
        for i, measure_i in enumerate(sim_distribution):
            for j, measure_j in enumerate(sim_distribution[measure_i]):
                var_lst.append(
                    np.var(np.array(sim_distribution[measure_i][measure_j]["sim"]))
                )
        E_VAR = np.mean(np.array(var_lst))

        # E(E(sim)) and VAR(E(sim)) across subj
        avg_lst = []
        for i, measure_i in enumerate(sim_distribution):
            for j, measure_j in enumerate(sim_distribution[measure_i]):
                avg_lst.append(
                    np.mean(np.array(sim_distribution[measure_i][measure_j]["sim"]))
                )
        E_E = np.mean(np.array(avg_lst))
        VAR_E = np.var(np.array(avg_lst))

        # change default_values to session_Rest1_LR
        if filter == "default_values":
            new_filter = "session_Rest1_LR"
            RESULTS[new_filter] = {}
            RESULTS[new_filter]["avg_mat"] = all_subjs_avg
            RESULTS[new_filter]["var_mat"] = across_subj_var
            RESULTS[new_filter]["avg_div_var_mat"] = np.divide(
                all_subjs_avg,
                np.sqrt(across_subj_var),
                out=np.zeros_like(all_subjs_avg),
                where=np.sqrt(across_subj_var) != 0,
            )
            RESULTS[new_filter]["sim_distribution"] = sim_distribution
            RESULTS[new_filter]["overall_stat"] = {
                "E_VAR": E_VAR,
                "E_E": E_E,
                "VAR_E": VAR_E,
            }
            RESULTS[new_filter]["name_lst"] = measure_name_lst
        else:
            RESULTS[filter] = {}
            RESULTS[filter]["avg_mat"] = all_subjs_avg
            RESULTS[filter]["var_mat"] = across_subj_var
            RESULTS[filter]["avg_div_var_mat"] = np.divide(
                all_subjs_avg,
                np.sqrt(across_subj_var),
                out=np.zeros_like(all_subjs_avg),
                where=np.sqrt(across_subj_var) != 0,
            )
            RESULTS[filter]["sim_distribution"] = sim_distribution
            RESULTS[filter]["overall_stat"] = {"E_VAR": E_VAR, "E_E": E_E, "VAR_E": VAR_E}
            RESULTS[filter]["name_lst"] = measure_name_lst

    ALL_RESULTS["dFC_similarity_overall"][metric] = deepcopy(RESULTS)

################# Hierarchical Clstr with Confidence Interval #################
"""
    - plot the most frequent hierclstr structures across subjects
"""
metric = "spearman"

for filter in ["default_values"]:

    Z_lst = list()
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        # SUBJs_output[filter]['all']['metric'] = (1, method, method)
        sim_mat = np.squeeze(SUBJs_output[filter]["all"][metric])

        dist_mat = corr2distance(sim_mat, metric=metric)
        Z = distance2Z(dist_mat, method="ward")
        Z_lst.append(Z)

ALL_RESULTS["Hierclstr_CI"] = deepcopy(Z_lst)

################# session ANOVA #################
"""
    - performing a two-way ANOVA test to investigate the effect of direction (LR/RL)
    and session on the observed similarity values
"""
sim_RESULTS = ALL_RESULTS["dFC_similarity_overall"]["spearman"]
# sim_RESULTS[filter]['sim_distribution'][measure_i][measure_j]['sim']

data = {}
for filter in sim_RESULTS:
    # session_Rest1_LR
    session = filter[filter.index("Rest") : filter.index("Rest") + 5]  # session_Rest1_LR
    direction = filter[filter.index("Rest") + 6 : filter.index("Rest") + 8]

    sim_distribution_all = sim_RESULTS[filter]["sim_distribution"]
    for i, measure_i in enumerate(sim_distribution_all):
        for j, measure_j in enumerate(sim_distribution_all[measure_i]):
            sim_distribution = sim_distribution_all[measure_i][measure_j]

            if not measure_i in data:
                data[measure_i] = {}
            if not measure_j in data[measure_i]:
                data[measure_i][measure_j] = {
                    "sim": list(),
                    "session": list(),
                    "direction": list(),
                }

            data[measure_i][measure_j]["sim"].extend(sim_distribution["sim"])
            data[measure_i][measure_j]["session"].extend(
                [session for sim in sim_distribution["sim"]]
            )
            data[measure_i][measure_j]["direction"].extend(
                [direction for sim in sim_distribution["sim"]]
            )

RESULTS = {}
for i, measure_i in enumerate(data):
    for j, measure_j in enumerate(data[measure_i]):

        if not measure_i in RESULTS:
            RESULTS[measure_i] = {}
        RESULTS[measure_i][measure_j] = two_way_anova(data[measure_i][measure_j])

ALL_RESULTS["session_ANOVA"] = deepcopy(RESULTS)

################# feature-based #################
"""
    - spatial
    - temporal
    - inter-time-correlation
    - inter-connection-correlation
    - dFC-avg
    - dFC-var
"""
ALL_RESULTS["dFC_similarity_feature_based"] = {}

feature2extract_list = [
    "spatial",
    "temporal",
    "inter_time_corr",
    "inter_conn_corr",
    "dFC_avg",
    "dFC_var",
]
for feature2extract in feature2extract_list:

    RESULTS = {}
    for filter in FILTERS:

        all_subjs_sim_mat = list()
        for s in ALL_RECORDS:
            SUBJs_output = np.load(
                assessment_results_root + FOLDER_name + s, allow_pickle="True"
            ).item()
            # SUBJs_output[filter]['feature_based'][feature2extract] = (sample, method, method)
            if np.any(
                np.isnan(
                    np.mean(
                        SUBJs_output[filter]["feature_based"][feature2extract], axis=0
                    )
                )
            ):
                continue
            all_subjs_sim_mat.append(
                np.mean(SUBJs_output[filter]["feature_based"][feature2extract], axis=0)
            )

        measure_name_lst = [
            measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
        ]
        all_subjs_sim_mat = np.array(all_subjs_sim_mat)
        all_subjs_avg = np.mean(all_subjs_sim_mat, axis=0)
        across_subj_var = np.var(all_subjs_sim_mat, axis=0)

        RESULTS[filter] = {}
        RESULTS[filter]["avg_mat"] = all_subjs_avg
        RESULTS[filter]["var_mat"] = across_subj_var
        RESULTS[filter]["avg_div_var_mat"] = np.divide(
            all_subjs_avg,
            np.sqrt(across_subj_var),
            out=np.zeros_like(all_subjs_avg),
            where=np.sqrt(across_subj_var) != 0,
        )
        RESULTS[filter]["name_lst"] = measure_name_lst

    ALL_RESULTS["dFC_similarity_feature_based"][feature2extract] = deepcopy(RESULTS)

############ Spatial vs. Temporal Scatter plot ############

for filter in ["default_values"]:

    all_subjs_spatial_sim_mat = list()
    all_subjs_temporal_sim_mat = list()
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        # SUBJs_output[filter]['feature_based'][feature2extract] = (sample, method, method)
        if np.any(
            np.isnan(np.mean(SUBJs_output[filter]["feature_based"]["spatial"], axis=0))
        ) or np.any(
            np.isnan(np.mean(SUBJs_output[filter]["feature_based"]["temporal"], axis=0))
        ):
            continue
        all_subjs_spatial_sim_mat.append(
            np.mean(SUBJs_output[filter]["feature_based"]["spatial"], axis=0)
        )
        all_subjs_temporal_sim_mat.append(
            np.mean(SUBJs_output[filter]["feature_based"]["temporal"], axis=0)
        )

    measure_name_lst = [
        measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
    ]
    all_subjs_spatial_sim_mat = np.mean(np.array(all_subjs_spatial_sim_mat), axis=0)
    all_subjs_temporal_sim_mat = np.mean(np.array(all_subjs_temporal_sim_mat), axis=0)

    scatter_data = {"spatial": list(), "temporal": list(), "labels": list()}
    for i in range(len(measure_name_lst)):
        for j in range(i):
            scatter_data["spatial"].append(all_subjs_spatial_sim_mat[i, j])
            scatter_data["temporal"].append(all_subjs_temporal_sim_mat[i, j])
            scatter_data["labels"].append(
                zip_name(measure_name_lst[i]) + "-" + zip_name(measure_name_lst[j])
            )

ALL_RESULTS["spatial_vs_temporal_similarity"] = deepcopy(scatter_data)

################# graph-based #################
"""
    - spatial
    - temporal-avg
    - ECM, shortest_path, degree, clustering_coef
"""

ALL_RESULTS["dFC_similarity_graph"] = {}

graph_property_list = ["ECM", "shortest_path", "degree", "clustering_coef"]
###### spatial #####
ALL_RESULTS["dFC_similarity_graph"]["spatial"] = {}

for graph_property in graph_property_list:

    RESULTS = {}
    for filter in FILTERS:

        all_subjs_sim_mat = list()
        for s in ALL_RECORDS:
            SUBJs_output = np.load(
                assessment_results_root + FOLDER_name + s, allow_pickle="True"
            ).item()
            # SUBJs_output[filter]['graph_based']['graph_spatial'][graph_property] = (sample, method, method)
            all_subjs_sim_mat.append(
                np.mean(
                    SUBJs_output[filter]["graph_based"]["graph_spatial"][graph_property],
                    axis=0,
                )
            )

        measure_name_lst = [
            measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
        ]
        all_subjs_sim_mat = np.array(all_subjs_sim_mat)
        all_subjs_avg = np.mean(all_subjs_sim_mat, axis=0)
        across_subj_var = np.var(all_subjs_sim_mat, axis=0)

        RESULTS[filter] = {}
        RESULTS[filter]["avg_mat"] = all_subjs_avg
        RESULTS[filter]["var_mat"] = across_subj_var
        RESULTS[filter]["avg_div_var_mat"] = np.divide(
            all_subjs_avg,
            np.sqrt(across_subj_var),
            out=np.zeros_like(all_subjs_avg),
            where=np.sqrt(across_subj_var) != 0,
        )
        RESULTS[filter]["name_lst"] = measure_name_lst

    ALL_RESULTS["dFC_similarity_graph"]["spatial"][graph_property] = deepcopy(RESULTS)

################################# TSNE #################################
"""
    - visualize subjects dFC obtained by different methods in a 2D space
        considering their distances
"""
RESULTS = {}
for filter in ["default_values"]:
    features_dict = {}
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        subj_name = SUBJs_output[filter]["TS_info_lst"][0]["subj_id_lst"][0]

        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):
            measure_name = measure_i.measure_name

            dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]
            # rank normalization
            dFC_mat_i = rank_norm(dFC_mat_i)

            if not measure_name in features_dict:
                features_dict[measure_name] = {"dFC": list(), "sample_name": list()}
            features_dict[measure_name]["dFC"].append(dFC_mat2vec(dFC_mat_i).flatten())
            features_dict[measure_name]["sample_name"].append(
                subj_name + "_" + measure_name
            )

sample_lst = list()
features_all = list()
for measure_name in features_dict:
    features_all.append(features_dict[measure_name]["dFC"])
    sample_lst.extend(features_dict[measure_name]["sample_name"])
features_all = np.array(features_all)

# observation matrix
X = np.reshape(features_all, (-1, features_all.shape[-1]))

RESULTS["sample_lst"] = sample_lst
# overall corr
RESULTS["corr"] = np.corrcoef(X)

RESULTS["X_red"] = {}
RESULTS["X_embedded"] = {}
RESULTS["X_red_corr"] = {}
for n_components in [100, 500, 1000, 1500, 2000]:
    pca = PCA(n_components=n_components)
    RESULTS["X_red"][n_components] = pca.fit_transform(X)

    # corr of subjects after PCA
    RESULTS["X_red_corr"][n_components] = np.corrcoef(RESULTS["X_red"][n_components])

    # embessing X into 2D space using TSNE
    RESULTS["X_embedded"][n_components] = TSNE(
        n_components=2, learning_rate="auto", init="random", perplexity=30
    ).fit_transform(RESULTS["X_red"][n_components])

ALL_RESULTS["TSNE"] = deepcopy(RESULTS)

################################# inter_subject similarity #################################

"""
    - returns correspondence of inter-subject relation between results of dFC
        measures in each session
    - dendogram based on inter-subject similarity
"""
ALL_RESULTS["subj_clustring"] = {}

subj_lvl_feature_lst = ["dFC_values", "FO", "dFC_var"]
for subj_lvl_feature in subj_lvl_feature_lst:
    RESULTS = {}
    inter_subj_sim_sessions = list()
    session_name_lst = list()
    for filter in FILTERS:
        if filter == "default_values":
            session_name_lst.append("session_Rest1_LR")
        elif "session" in filter:
            session_name_lst.append(filter)
        features_subj = list()
        inter_subj_sim_session = list()
        subj_name_lst = list()
        for s in ALL_RECORDS:
            SUBJs_output = np.load(
                assessment_results_root + FOLDER_name + s, allow_pickle="True"
            ).item()
            subj_name_lst.append(SUBJs_output[filter]["TS_info_lst"][0]["subj_id_lst"][0])

            features_i = list()
            for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

                if subj_lvl_feature == "dFC_values":
                    dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]
                    # rank normalization
                    dFC_mat_i = rank_norm(dFC_mat_i)
                    features_i.append(dFC_mat2vec(dFC_mat_i).flatten())
                elif subj_lvl_feature == "dFC_var":
                    dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]
                    # rank normalization
                    # dFC_mat_i = rank_norm(dFC_mat_i)
                    dFC_var = np.var(dFC_mat_i, axis=0)
                    dFC_var = rank_norm(dFC_var)
                    dFC_var = cat_data(dFC_var, N=10)
                    dFC_var = np.where(dFC_var == np.max(dFC_var), 1, 0)
                    features_i.append(dFC_mat2vec(dFC_var))
                elif subj_lvl_feature == "FO":
                    FO = SUBJs_output[filter]["FO"][i]
                    if not measure_i.is_state_based:
                        continue
                    else:
                        FO = [FO[FCS] for FCS in FO]
                        FO = np.array(FO)
                    features_i.append(FO)
            features_i = np.array(features_i)
            features_subj.append(features_i)

        features_subj = np.array(
            features_subj
        )  # features_subj = (subj, method, subj_lvl_feature)
        num_subj = features_subj.shape[0]

        for i in range(features_subj.shape[1]):

            features_i = np.squeeze(
                features_subj[:, i, :]
            )  # features_i = (subj, subj_lvl_feature)

            inter_subj_sim_i = np.corrcoef(features_i)  # inter_subj_sim_i = (subj, subj)
            # inter_subj_sim_i, p_value = stats.spearmanr(features_i, axis=1) # inter_subj_sim_i = (subj, subj)

            inter_subj_sim_i = dFC_mat2vec(inter_subj_sim_i)

            inter_subj_sim_session.append(inter_subj_sim_i)

        inter_subj_sim_session = np.array(
            inter_subj_sim_session
        )  # (method, inter_subj_sim_values)

        inter_subj_sim_sessions.append(inter_subj_sim_session)

    inter_subj_sim_sessions = np.array(
        inter_subj_sim_sessions
    )  # (session, method, inter_subj_sim_values)

    if subj_lvl_feature == "dFC_values":
        measure_name_lst = [
            measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
        ]
    elif subj_lvl_feature == "FO":
        measure_name_lst = [
            measure.measure_name
            for measure in SUBJs_output[filter]["measure_lst"]
            if measure.is_state_based
        ]

    RESULTS["across_method"] = {}
    for session_id, session in enumerate(session_name_lst):
        sim_mat = np.zeros((len(measure_name_lst), len(measure_name_lst)))
        for i in range(len(measure_name_lst)):
            for j in range(len(measure_name_lst)):

                spear_coef, p_value = stats.spearmanr(
                    inter_subj_sim_sessions[session_id, i, :],
                    inter_subj_sim_sessions[session_id, j, :],
                )
                sim_mat[i, j] = spear_coef

        RESULTS["across_method"][session] = {}
        RESULTS["across_method"][session]["sim_mat"] = sim_mat
        RESULTS["across_method"][session]["name_lst"] = measure_name_lst

    ALL_RESULTS["subj_clustring"][subj_lvl_feature] = deepcopy(RESULTS)

################################# dFC var #################################

"""
    - avg(variance/fluctuations of dFC in one Subj)
    - rank normed
"""
for filter in ["default_values"]:
    RESULTS = {}
    RESULTS["avg_dFC_var"] = {}
    RESULTS["var_dFC_var"] = {}
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        node_networks = node_info2network(
            SUBJs_output[filter]["TS_info_lst"][0]["nodes_info"]
        )
        n_regions = SUBJs_output[filter]["TS_info_lst"][0]["n_regions"]

        for i, measure in enumerate(SUBJs_output[filter]["measure_lst"]):
            if not measure.measure_name in RESULTS["avg_dFC_var"]:
                RESULTS["avg_dFC_var"][measure.measure_name] = list()
            # SUBJs_output[filter]['dFC_var'][i] = (1, connection)
            var_mat = np.squeeze(
                dFC_vec2mat(SUBJs_output[filter]["dFC_var"][i], N=n_regions)
            )  # (ROI, ROI)
            np.fill_diagonal(var_mat, 0)
            RESULTS["avg_dFC_var"][measure.measure_name].append(rank_norm(var_mat))

    for key in RESULTS["avg_dFC_var"]:
        RESULTS["avg_dFC_var"][key] = np.array(RESULTS["avg_dFC_var"][key])
        RESULTS["var_dFC_var"][key] = np.var(RESULTS["avg_dFC_var"][key], axis=0)
        RESULTS["avg_dFC_var"][key] = np.mean(RESULTS["avg_dFC_var"][key], axis=0)

ALL_RESULTS["dFC_var"] = deepcopy(RESULTS)

################################# dFC avg #################################

"""
    - avg(avg of dFC -static FC- in one Subj)
    - rank normed
"""

for filter in ["default_values"]:
    RESULTS = {}
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        node_networks = node_info2network(
            SUBJs_output[filter]["TS_info_lst"][0]["nodes_info"]
        )
        n_regions = SUBJs_output[filter]["TS_info_lst"][0]["n_regions"]

        for i, measure in enumerate(SUBJs_output[filter]["measure_lst"]):
            if not measure.measure_name in RESULTS:
                RESULTS[measure.measure_name] = list()
            # SUBJs_output[filter]['dFC_avg'][i] = (1, connection)
            avg_mat = np.squeeze(
                dFC_vec2mat(SUBJs_output[filter]["dFC_avg"][i], N=n_regions)
            )  # (ROI, ROI)
            RESULTS[measure.measure_name].append(rank_norm(avg_mat))

    for key in RESULTS:
        RESULTS[key] = np.array(RESULTS[key])
        RESULTS[key] = np.mean(RESULTS[key], axis=0)

ALL_RESULTS["dFC_avg"] = deepcopy(RESULTS)

################################# Across Func Conn total Correlation #################################

"""
    - spearman_corr((dFConnection(node_i, node_j) timecourse using method m), (dFConnection(node_i, node_j) timecourse using method n))
"""

for filter in ["default_values"]:
    RESULTS = {}
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        node_networks = node_info2network(
            SUBJs_output[filter]["TS_info_lst"][0]["nodes_info"]
        )
        n_regions = SUBJs_output[filter]["TS_info_lst"][0]["n_regions"]
        n_time = SUBJs_output[filter]["dFCM_samples"][str(0)].shape[0]

        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

            dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]
            # rank normalization
            dFC_mat_i_norm = stats.rankdata(dFC_mat_i.flatten()).reshape(
                n_time, n_regions, n_regions
            )
            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i_norm)

            for j, measure_j in enumerate(SUBJs_output[filter]["measure_lst"]):

                if j >= i:
                    continue

                dFC_mat_j = SUBJs_output[filter]["dFCM_samples"][str(j)]
                # rank normalization
                dFC_mat_j_norm = stats.rankdata(dFC_mat_j.flatten()).reshape(
                    n_time, n_regions, n_regions
                )
                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j_norm)

                sim = list()
                for func_conn in range(dFC_mat_i_vec.shape[1]):
                    if (
                        np.var(dFC_mat_i_vec[:, func_conn]) == 0
                        or np.var(dFC_mat_j_vec[:, func_conn]) == 0
                    ):
                        sim.append(0)
                    else:
                        sim.append(
                            np.corrcoef(
                                dFC_mat_i_vec[:, func_conn], dFC_mat_j_vec[:, func_conn]
                            )[0, 1]
                        )

                sim = np.array(sim)
                if not measure_i.measure_name in RESULTS:
                    RESULTS[measure_i.measure_name] = {}
                if not measure_j.measure_name in RESULTS[measure_i.measure_name]:
                    RESULTS[measure_i.measure_name][measure_j.measure_name] = list()

                RESULTS[measure_i.measure_name][measure_j.measure_name].append(
                    dFC_vec2mat(sim[None, :], N=n_regions)[0]
                )

    measure_lst = [
        measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
    ]
    for key_i in RESULTS:
        for key_j in RESULTS[key_i]:
            RESULTS[key_i][key_j] = np.mean(np.array(RESULTS[key_i][key_j]), axis=0)

ALL_RESULTS["across_func_conns"] = deepcopy(RESULTS)

################################# High Variation Regions #################################
"""
    - high variation regions over methods and over time.
"""
for filter in ["default_values"]:
    var_over_time = list()
    var_over_method = list()
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        node_networks = node_info2network(
            SUBJs_output[filter]["TS_info_lst"][0]["nodes_info"]
        )

        dFC_mat_lst = list()
        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

            dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]

            # rank normalization
            dFC_mat_i = rank_norm(dFC_mat_i)

            # dFC mat
            dFC_mat_lst.append(dFC_mat_i)

        dFC_mat_lst = np.array(dFC_mat_lst)  # (method, time, ROI, ROI)
        var_over_method.append(np.mean(np.var(dFC_mat_lst, axis=0), axis=0))
        var_over_time.append(np.mean(np.var(dFC_mat_lst, axis=1), axis=0))

    var_over_time = np.array(var_over_time)  # (subj, ROI, ROI)
    var_over_method = np.array(var_over_method)  # (subj, ROI, ROI)

    # collect var over method and time across all func conns of all subjects
    scatter_data = {"var_method": list(), "var_time": list()}
    scatter_data["var_method"] = dFC_mat2vec(
        var_over_method
    ).flatten()  # (subj*(ROI)*(ROI-1)/2,)
    scatter_data["var_time"] = dFC_mat2vec(
        var_over_time
    ).flatten()  # (subj*(ROI)*(ROI-1)/2,)

    var_over_time = np.mean(var_over_time, axis=0)  # (ROI, ROI)
    var_over_method = np.mean(var_over_method, axis=0)  # (ROI, ROI)

    RESULTS = {}
    RESULTS["var_over_time"] = var_over_time
    RESULTS["var_over_method"] = var_over_method

    RATIO = {
        "var across method / time - 1": np.divide(
            var_over_method,
            var_over_time,
            out=np.zeros_like(var_over_method),
            where=var_over_time != 0,
        )
        - 1
    }

    ALL_RESULTS["var_across_func_conns"] = deepcopy(RESULTS)
    ALL_RESULTS["var_across_func_conns_ratio"] = deepcopy(RATIO)
    ALL_RESULTS["var_method_vs_time_across_func_conns_scatter"] = deepcopy(scatter_data)

    for key in RESULTS:
        RESULTS[key] = rank_norm(RESULTS[key])

    for key in RESULTS:
        RESULTS[key] = cat_data(RESULTS[key], N=10)
        RESULTS[key] = np.where(RESULTS[key] == np.max(RESULTS[key]), 1, 0)

    ALL_RESULTS["high_var_func_conns"] = deepcopy(RESULTS)

################################# Variation Value Comparison #################################
"""
    - compare variation over methods with variation over time.
"""
RESULTS = {}
for filter in ["default_values"]:
    diff_mat_dict = {}
    temp_var_dict = {}
    lag_1_diff_mat_dict = {}
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):
            measure_name_i = measure_i.measure_name

            dFC_mat_i = rank_norm(SUBJs_output[filter]["dFCM_samples"][str(i)])

            for j, measure_j in enumerate(SUBJs_output[filter]["measure_lst"]):
                measure_name_j = measure_j.measure_name

                dFC_mat_j = rank_norm(SUBJs_output[filter]["dFCM_samples"][str(j)])

                if not measure_name_i in diff_mat_dict:
                    diff_mat_dict[measure_name_i] = {}
                    temp_var_dict[measure_name_i] = {}
                    lag_1_diff_mat_dict[measure_name_i] = {}
                if not measure_name_j in diff_mat_dict[measure_name_i]:
                    diff_mat_dict[measure_name_i][measure_name_j] = list()
                    temp_var_dict[measure_name_i][measure_name_j] = list()
                    lag_1_diff_mat_dict[measure_name_i][measure_name_j] = list()

                dFC = np.concatenate(
                    (dFC_mat_i[None, :, :, :], dFC_mat_j[None, :, :, :]), axis=0
                )
                diff_mat_dict[measure_name_i][measure_name_j].append(
                    np.mean(np.var(dFC, axis=0), axis=0)
                )
                temp_var_dict[measure_name_i][measure_name_j].append(
                    np.mean(np.var(dFC, axis=1), axis=0)
                )
                lag_1_diff_mat_dict[measure_name_i][measure_name_j].append(
                    np.mean(np.mean(0.25 * (np.diff(dFC, axis=1) ** 2), axis=1), axis=0)
                )

    measure_name_lst = [measure_key_i for measure_key_i in diff_mat_dict]

    scatter_data = {"var_method": list(), "var_time": list(), "labels": list()}
    scatter_data_across_func_conn = {}
    divide_temp = np.zeros((len(measure_name_lst), len(measure_name_lst)))
    divide_1_lag = np.zeros((len(measure_name_lst), len(measure_name_lst)))
    for i, measure_key_i in enumerate(measure_name_lst):
        for j, measure_key_j in enumerate(measure_name_lst):

            A = np.mean(np.array(diff_mat_dict[measure_key_i][measure_key_j]), axis=0)
            B = np.mean(np.array(temp_var_dict[measure_key_i][measure_key_j]), axis=0)
            C = np.mean(
                np.array(lag_1_diff_mat_dict[measure_key_i][measure_key_j]), axis=0
            )

            # collect data for scatter plot
            if j < i:
                if not zip_name(measure_key_i) in scatter_data_across_func_conn:
                    scatter_data_across_func_conn[zip_name(measure_key_i)] = {}
                if (
                    not zip_name(measure_key_j)
                    in scatter_data_across_func_conn[zip_name(measure_key_i)]
                ):
                    scatter_data_across_func_conn[zip_name(measure_key_i)][
                        zip_name(measure_key_j)
                    ] = {"var_method": list(), "var_time": list()}

                scatter_data_across_func_conn[zip_name(measure_key_i)][
                    zip_name(measure_key_j)
                ]["var_method"] = A.flatten()
                scatter_data_across_func_conn[zip_name(measure_key_i)][
                    zip_name(measure_key_j)
                ]["var_time"] = B.flatten()

                scatter_data["var_method"].append(np.mean(A))
                scatter_data["var_time"].append(np.mean(B))
                scatter_data["labels"].append(
                    zip_name(measure_key_i) + "-" + zip_name(measure_key_j)
                )

            divide_temp[i, j] = np.mean(
                np.divide(A, B, out=np.zeros_like(A), where=B != 0)
            )
            divide_1_lag[i, j] = np.mean(
                np.divide(A, C, out=np.zeros_like(A), where=C != 0)
            )

    RESULTS["var_method_divide_temp"] = {
        "sim_mat": divide_temp,
        "name_lst": measure_name_lst,
    }
    RESULTS["var_method_divide_1_lag"] = {
        "sim_mat": divide_1_lag,
        "name_lst": measure_name_lst,
    }

ALL_RESULTS["var_comparison"] = deepcopy(RESULTS)
ALL_RESULTS["var_method_vs_time_method_pairs_across_func_conns"] = deepcopy(
    scatter_data_across_func_conn
)
ALL_RESULTS["var_method_vs_time_method_pairs"] = deepcopy(scatter_data)

################################# Var method vs. Time Clstrwise #################################
"""
    - compute var over method and time within each group of methods
"""
ALL_RESULTS["var_method_vs_time_clstrwise"] = {}

clstr1 = ["CAP", "Windowless"]
clstr2 = ["Clustering", "ContinuousHMM", "DiscreteHMM"]
clstr3 = ["SlidingWindow", "Time-Freq"]
clstrs_dict = {"clstr1": clstr1, "clstr2": clstr2, "clstr3": clstr3}

for filter in ["default_values"]:
    var_over_time = {"clstr1": list(), "clstr2": list(), "clstr3": list()}
    var_over_method = {"clstr1": list(), "clstr2": list(), "clstr3": list()}
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        dFC_mat_lst = {"clstr1": list(), "clstr2": list(), "clstr3": list()}
        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

            dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]

            # rank normalization
            dFC_mat_i = rank_norm(dFC_mat_i)

            # dFC mat
            if measure_i.measure_name in clstr1:
                dFC_mat_lst["clstr1"].append(dFC_mat_i)
            elif measure_i.measure_name in clstr2:
                dFC_mat_lst["clstr2"].append(dFC_mat_i)
            elif measure_i.measure_name in clstr3:
                dFC_mat_lst["clstr3"].append(dFC_mat_i)

        # Calc var_over_method and var_over_time in each subj
        for clstr in dFC_mat_lst:
            dFC_mat_lst[clstr] = np.array(dFC_mat_lst[clstr])  # (method, time, ROI, ROI)
            var_over_method[clstr].append(
                np.mean(np.var(dFC_mat_lst[clstr], axis=0), axis=0)
            )
            var_over_time[clstr].append(
                np.mean(np.var(dFC_mat_lst[clstr], axis=1), axis=0)
            )

    for clstr in dFC_mat_lst:
        var_over_time[clstr] = np.array(var_over_time[clstr])  # (subj, ROI, ROI)
        var_over_method[clstr] = np.array(var_over_method[clstr])  # (subj, ROI, ROI)

        var_over_time[clstr] = dFC_mat2vec(
            var_over_time[clstr]
        ).flatten()  # (subj*(ROI)*(ROI-1)/2,)
        var_over_method[clstr] = dFC_mat2vec(
            var_over_method[clstr]
        ).flatten()  # (subj*(ROI)*(ROI-1)/2,)

    # collect var over method and time across all func conns of all subjects
    scatter_data = {}
    for clstr in var_over_method:
        scatter_data[clstr] = {"var_method": list(), "var_time": list()}
        scatter_data[clstr]["var_method"] = var_over_method[clstr]
        scatter_data[clstr]["var_time"] = var_over_time[clstr]

    ALL_RESULTS["var_method_vs_time_clstrwise"]["scatter_data"] = deepcopy(scatter_data)
    ALL_RESULTS["var_method_vs_time_clstrwise"]["clstrs_dict"] = deepcopy(clstrs_dict)

################################# Inter-subject Variation #################################
"""
    - compute var over method and subject across all func conns
"""
for filter in ["default_values"]:
    all_dFC = list()
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        node_networks = node_info2network(
            SUBJs_output[filter]["TS_info_lst"][0]["nodes_info"]
        )

        dFC_mat_lst = list()
        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

            dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]

            # rank normalization
            dFC_mat_i = rank_norm(dFC_mat_i)

            # dFC mat
            dFC_mat_lst.append(dFC_mat_i)

        dFC_mat_lst = np.array(dFC_mat_lst)  # (method, time, ROI, ROI)
        all_dFC.append(dFC_mat_lst)

    all_dFC = np.array(all_dFC)  # (subj, method, time, ROI, ROI)
    var_over_subj = np.mean(
        np.mean(np.var(all_dFC, axis=0), axis=0), axis=0
    )  # (ROI, ROI)
    var_over_method = np.mean(
        np.mean(np.var(all_dFC, axis=1), axis=0), axis=0
    )  # (ROI, ROI)

    # collect var over method and time across all func conns of all subjects
    scatter_data = {
        "var_method": dFC_mat2vec(var_over_method).flatten(),  # ((ROI)*(ROI-1)/2,)
        "var_subj": dFC_mat2vec(var_over_subj).flatten(),  # ((ROI)*(ROI-1)/2,)
    }

    RESULTS = {}
    RESULTS["var_over_subj"] = var_over_subj
    RESULTS["var_over_method"] = var_over_method

    RATIO = {
        "var across method / subj - 1": np.divide(
            var_over_method,
            var_over_subj,
            out=np.zeros_like(var_over_method),
            where=var_over_subj != 0,
        )
        - 1
    }

    ALL_RESULTS["inter-subj_var_across_func_conns"] = deepcopy(RESULTS)
    ALL_RESULTS["inter-subj_var_across_func_conns_ratio"] = deepcopy(RATIO)
    ALL_RESULTS["var_method_vs_subj_across_func_conns_scatter"] = deepcopy(scatter_data)

################################# Overall Variability #################################
"""
    - compute var over method and overall across all func conns
"""

for filter in ["default_values"]:
    all_dFC = list()
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        node_networks = node_info2network(
            SUBJs_output[filter]["TS_info_lst"][0]["nodes_info"]
        )

        dFC_mat_lst = list()
        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

            dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]

            # rank normalization
            dFC_mat_i = rank_norm(dFC_mat_i)

            # dFC mat
            dFC_mat_lst.append(dFC_mat_i)

        dFC_mat_lst = np.array(dFC_mat_lst)  # (method, time, ROI, ROI)
        all_dFC.append(dFC_mat_lst)

    all_dFC = np.array(all_dFC)  # (subj, method, time, ROI, ROI)
    var_over_method = np.mean(
        np.mean(np.var(all_dFC, axis=1), axis=0), axis=0
    )  # (ROI, ROI)
    all_dFC = np.transpose(all_dFC, (1, 0, 2, 3, 4))  # (method, subj, time, ROI, ROI)
    all_dFC = all_dFC.reshape(
        all_dFC.shape[0], -1, all_dFC.shape[3], all_dFC.shape[4]
    )  # (method, subj*time, ROI, ROI)
    var_over_all = np.mean(np.var(all_dFC, axis=1), axis=0)  # (ROI, ROI)

    # collect var over method and time across all func conns of all subjects
    scatter_data = {
        "var_method": dFC_mat2vec(var_over_method).flatten(),  # ((ROI)*(ROI-1)/2,)
        "var_all": dFC_mat2vec(var_over_all).flatten(),  # ((ROI)*(ROI-1)/2,)
    }

    RESULTS = {}
    RESULTS["var_over_all"] = var_over_all
    RESULTS["var_over_method"] = var_over_method

    RATIO = {
        "var across method / all - 1": np.divide(
            var_over_method,
            var_over_all,
            out=np.zeros_like(var_over_method),
            where=var_over_all != 0,
        )
        - 1
    }

    ALL_RESULTS["var_all_across_func_conns"] = deepcopy(RESULTS)
    ALL_RESULTS["var_all_across_func_conns_ratio"] = deepcopy(RATIO)
    ALL_RESULTS["var_method_vs_all_across_func_conns_scatter"] = deepcopy(scatter_data)

################################# Randomization Tests #################################

metric = "spearman"

ALL_RESULTS["randomization"] = {}

########### Similarity with static FC ###########
"""
find the similarity between the dFC obtained by each method
and a dFC which is a constant sequence of mean of all methods dFC
"""
for filter in ["default_values"]:

    RESULTS = {}
    RESULTS["sim"] = list()
    RESULTS["dFC_method"] = list()
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        dFC_dict = {}
        for i, measure in enumerate(SUBJs_output[filter]["measure_lst"]):
            dFC_mat = SUBJs_output[filter]["dFCM_samples"][str(i)]
            dFC_dict[measure.measure_name] = dFC_mat

        dFC_mean = list()
        for i, measure_name in enumerate(dFC_dict):
            dFC_mat = dFC_dict[measure_name]
            n_time = dFC_mat.shape[0]
            dFC_mean.append(np.mean(dFC_mat, axis=0))
        dFC_mean = np.mean(np.array(dFC_mean), axis=0)
        dFC_mean = np.repeat(dFC_mean[None, :, :], n_time, axis=0)
        dFC_mean_vec = dFC_mat2vec(dFC_mean)

        for i, measure_i_name in enumerate(dFC_dict):

            dFC_mat_i = dFC_dict[measure_i_name]
            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mean_vec.flatten())

            RESULTS["sim"].append(sim)
            RESULTS["dFC_method"].append(measure_i_name)

ALL_RESULTS["randomization"]["sim_with_static_FC"] = deepcopy(RESULTS)

########### Shuffled ###########

"""
find the similarity between the dFC obtained by each method
but with randomized temporal(time)/spatial(region)/all(both) order
 - temporal: find the similarity between the dFC obtained by each method
            but with randomized temporal order. The null keeps the spatial
            order but not the temporal orders
 - spatial: find the similarity between the dFC obtained by each method
            but with randomized spatial order. The null keeps the temporal
            order but not the spatial orders
 - temporal: find the similarity between the dFC obtained by each method
            but with randomized both temporal AND spatial order. The null keeps
            neither the spatial order nor the temporal order
"""
metric = "spearman"
modes_lst = ["temporal", "spatial", "all"]
ALL_RESULTS["randomization"]["shuffled"] = {"similarity_dict": {}, "modes_lst": modes_lst}

for filter in ["default_values"]:

    all_subjs_sim_mat = list()
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        # SUBJs_output[filter]['all']['metric'] = (1, method, method)
        all_subjs_sim_mat.append(np.squeeze(SUBJs_output[filter]["all"][metric]))

    all_subjs_sim_mat = np.array(all_subjs_sim_mat)
    all_subjs_avg = np.mean(all_subjs_sim_mat, axis=0)
    measure_name_lst = [
        measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
    ]

    RESULTS = {}
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        dFC_dict = {}
        for i, measure in enumerate(SUBJs_output[filter]["measure_lst"]):

            dFC_mat = SUBJs_output[filter]["dFCM_samples"][str(i)]
            dFC_dict[zip_name(measure.measure_name)] = dFC_mat

        for mode in modes_lst:
            key = mode + "_shuffle_sim"
            output = randomized_dFC_sim(dFC_dict, N=num_randomization, mode=mode)

            for measure_i_name in output:
                for measure_j_name in output[measure_i_name]:

                    if not measure_i_name in RESULTS:
                        RESULTS[measure_i_name] = {}
                    if not measure_j_name in RESULTS[measure_i_name]:
                        RESULTS[measure_i_name][measure_j_name] = {}
                    if not key in RESULTS[measure_i_name][measure_j_name]:
                        RESULTS[measure_i_name][measure_j_name][key] = list()

                    RESULTS[measure_i_name][measure_j_name][key].extend(
                        output[measure_i_name][measure_j_name]["sim"]
                    )

    for measure_i_name in output:
        for measure_j_name in output[measure_i_name]:
            sim = all_subjs_avg[
                measure_name_lst.index(unzip_name(measure_i_name)),
                measure_name_lst.index(unzip_name(measure_j_name)),
            ]
            # change diagonal from 0 to 1
            if measure_i_name == measure_j_name:
                sim = 1
            RESULTS[measure_i_name][measure_j_name]["actual_sim"] = [
                sim for item in RESULTS[measure_i_name][measure_j_name][key]
            ]

ALL_RESULTS["randomization"]["shuffled"]["similarity_dict"] = RESULTS

########### Random state time course ###########

"""
find the similarity between dFC matrices created using
a random state time course but real FC patterns obtained
by each method; for SW and TF all FC patterns of each
subject are used
"""
for filter in ["default_values"]:

    all_subjs_sim_mat = list()
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        # SUBJs_output[filter]['all']['metric'] = (1, method, method)
        all_subjs_sim_mat.append(np.squeeze(SUBJs_output[filter]["all"][metric]))

    all_subjs_sim_mat = np.array(all_subjs_sim_mat)
    all_subjs_avg = np.mean(all_subjs_sim_mat, axis=0)
    measure_name_lst = [
        measure.measure_name for measure in SUBJs_output[filter]["measure_lst"]
    ]

    RESULTS = {}
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        n_time = SUBJs_output[filter]["dFCM_samples"][str(0)].shape[0]

        FCS_dict = {}
        for i, measure in enumerate(SUBJs_output[filter]["measure_lst"]):
            FCS = measure.FCS
            if len(FCS) == 0:
                FCS = SUBJs_output[filter]["dFCM_samples"][str(i)]
            FCS_dict[zip_name(measure.measure_name)] = FCS

        output = dFC_rand_sim(FCS_dict, n_time, N=num_randomization)

        for measure_i_name in output:
            for measure_j_name in output[measure_i_name]:

                if not measure_i_name in RESULTS:
                    RESULTS[measure_i_name] = {}
                if not measure_j_name in RESULTS[measure_i_name]:
                    RESULTS[measure_i_name][measure_j_name] = {"sim": list(), "": list()}

                RESULTS[measure_i_name][measure_j_name]["sim"].extend(
                    output[measure_i_name][measure_j_name]["sim"]
                )
                RESULTS[measure_i_name][measure_j_name][""].extend(
                    output[measure_i_name][measure_j_name][""]
                )

    for measure_i_name in output:
        for measure_j_name in output[measure_i_name]:
            sim = all_subjs_avg[
                measure_name_lst.index(unzip_name(measure_i_name)),
                measure_name_lst.index(unzip_name(measure_j_name)),
            ]
            # change diagonal from 0 to 1
            if measure_i_name == measure_j_name:
                sim = 1
            RESULTS[measure_i_name][measure_j_name]["actual_sim"] = [
                sim for item in RESULTS[measure_i_name][measure_j_name]["sim"]
            ]

ALL_RESULTS["randomization"]["random_state_TC"] = deepcopy(RESULTS)

################################# inter-state spatial similarity #################################

key_name = "inter-state_spatial_sim"
RESULTS = {"sim": list(), "dFC_method": list()}
for measure in ALL_RESULTS["measure_lst"]:

    if len(measure.FCS) == 0:
        continue
    FCSs = np.array(measure.FCS)

    for i, FCS_i in enumerate(FCSs):
        for j, FCS_j in enumerate(FCSs):

            if j >= i:
                continue

            vec_i = dFC_mat2vec(FCS_i)
            vec_j = dFC_mat2vec(FCS_j)

            RESULTS["sim"].append(np.corrcoef(vec_i, vec_j)[0, 1])
            RESULTS["dFC_method"].append(measure.measure_name)

ALL_RESULTS[key_name] = deepcopy(RESULTS)

################################# SIMILARITY OF ADJACENT TIME POINTS #################################

RESULTS = {}
key_name = "similarity of adjacent time points"
for filter in ["default_values"]:
    RESULTS[key_name] = list()
    RESULTS["dFC_method"] = list()
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()
        n_time = SUBJs_output[filter]["dFCM_samples"][str(0)].shape[0]

        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

            dFC_mat_i = SUBJs_output[filter]["dFCM_samples"][str(i)]

            for t in range(n_time - 1):
                sim_t, p = stats.spearmanr(
                    dFC_mat_i[t, :, :].flatten(), dFC_mat_i[t + 1, :, :].flatten()
                )

                RESULTS[key_name].append(sim_t)
                RESULTS["dFC_method"].append(measure_i.measure_name)

    ALL_RESULTS["adjacent_time_points"] = deepcopy(RESULTS)

################################# TRANSITION FREQUENCY #################################
"""
 - plot normalized transition frequency
"""
RESULTS = {}
key_name = "trans_freq"
for filter in ["default_values"]:
    RESULTS[key_name] = list()
    RESULTS["dFC_method"] = list()
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):

            if SUBJs_output[filter]["transition_stats"][i] == {}:
                continue
            RESULTS[key_name].append(
                SUBJs_output[filter]["transition_stats"][i]["trans_norm"]
            )
            RESULTS["dFC_method"].append(measure_i.measure_name)

    ALL_RESULTS["transition_freq"] = deepcopy(RESULTS)

################################# DWELL TIME #################################
"""
 - plot normalized (not downsampled) dwell times
 - the dwell times are not averaged within each subj, they include all individual
 dwells
"""
RESULTS = {}
key_name = "dwell_time"
for filter in ["default_values"]:
    RESULTS[key_name] = list()
    RESULTS["dFC_method"] = list()
    for s in ALL_RECORDS:

        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        for i, measure_i in enumerate(SUBJs_output[filter]["measure_lst"]):
            if SUBJs_output[filter]["transition_stats"][i] == {}:
                continue
            dwell_time_lst = SUBJs_output[filter]["transition_stats"][i][
                "dwell_time_norm"
            ]
            # add each of the ind dwell times to the list along with the method's name
            for dwell_time in dwell_time_lst:
                RESULTS[key_name].append(dwell_time)
                RESULTS["dFC_method"].append(measure_i.measure_name)

    ALL_RESULTS["dwell_time"] = deepcopy(RESULTS)

################################# TIME RECORD #################################

RESULTS = {}
for filter in ["default_values"]:

    RESULTS["FCS_fit_time (s)"] = list()
    RESULTS["dFC_assess_time (s)"] = list()
    RESULTS["dFC_method"] = list()
    for s in ALL_RECORDS:
        SUBJs_output = np.load(
            assessment_results_root + FOLDER_name + s, allow_pickle="True"
        ).item()

        for measure_id in SUBJs_output[filter]["time_record_dict"]:

            if SUBJs_output[filter]["time_record_dict"][measure_id]["FCS_fit"] is None:
                RESULTS["FCS_fit_time (s)"].append(0)
            else:
                RESULTS["FCS_fit_time (s)"].append(
                    SUBJs_output[filter]["time_record_dict"][measure_id]["FCS_fit"]
                )
            RESULTS["dFC_assess_time (s)"].append(
                SUBJs_output[filter]["time_record_dict"][measure_id]["dFC_assess"]
            )
            RESULTS["dFC_method"].append(
                SUBJs_output[filter]["measure_lst"][int(measure_id)].measure_name
            )

    ALL_RESULTS["time_record"] = deepcopy(RESULTS)


################################# SAVE ALL RESULTS #################################

np.save(output_root + "ALL_RESULTS.npy", ALL_RESULTS)

#################################################################################



"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import time

import numpy as np
from sklearn.cluster import KMeans

from ..dfc import DFC
from ..dfc_utils import KMeansCustom, dFC_mat2vec, dFC_vec2mat
from ..time_series import TIME_SERIES
from .base_dfc_method import BaseDFCMethod
from .sliding_window import SLIDING_WINDOW
from .time_freq import TIME_FREQ

########################### Sliding_Window + Clustering ############################

"""
- We used a tapered window as in Allen et al., created by convolving a rectangle (width = 22 TRs = 44s)
  with a Gaussian (σ = 3 TRs) and slid in steps of 1 TR, resulting in W= 126 windows (Allen et al., 2014).
- Kmeans Clustering is repeated 500 times to escape local minima (Allen et al., 2014)
- for clustering, we have a 2-level kmeans clustering. First, we cluster FCSs of each subject. Then, we
    cluster all clustering centers from all subjects. the final estimate_dFC is using the second kmeans
    model (Allen et al., 2014; Ou et al., 2015).

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.
"""


class SLIDING_WINDOW_CLUSTR(BaseDFCMethod):

    def __init__(self, **params):

        self.logs_ = ""
        self.TPM = []
        self.FCS_ = []
        self.mean_act = []
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None

        self.params_name_lst = [
            "measure_name",
            "is_state_based",
            "clstr_base_measure",
            "sw_method",
            "tapered_window",
            "clstr_distance",
            "coi_correction",
            "n_subj_clstrs",
            "W",
            "window_std",
            "n_overlap",
            "n_states",
            "normalization",
            "n_jobs",
            "verbose",
            "backend",
            "num_subj",
            "num_select_nodes",
            "num_time_point",
            "Fs_ratio",
            "noise_ratio",
            "num_realization",
            "session",
        ]
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params["measure_name"] = "Clustering"
        self.params["is_state_based"] = True

        assert (
            self.params["clstr_distance"] == "euclidean"
            or self.params["clstr_distance"] == "manhattan"
        ), "Clustering distance not recognized. It must be either \
                euclidean or manhattan."

        assert (
            self.params["clstr_base_measure"] in self.base_methods_name_lst
        ), "Base method not recognized."

    @property
    def measure_name(self):
        return self.params["measure_name"]  # + '_' + self.base_method

    def dFC_mat2vec(self, C_t):
        return dFC_mat2vec(C_t)

    def dFC_vec2mat(self, F, N):
        return dFC_vec2mat(F=F, N=N)

    def clusters_lst2idx(self, clusters):
        Z = np.zeros((self.F.shape[0],))
        for i, cluster in enumerate(clusters):
            for sample in cluster:
                Z[sample] = i
        return Z.astype(int)

    def cluster_FC(self, FCS_raw, n_clusters, n_regions):

        F = self.dFC_mat2vec(FCS_raw)

        if self.params["clstr_distance"] == "manhattan":
            ########### Manhattan Clustering ##############
            kmeans_ = KMeansCustom(
                n_clusters=n_clusters,
                n_init=500,
                init="k-means++",
                metric="manhattan",
            ).fit(F)
            kmeans_.cluster_centers_ = kmeans_.cluster_centers_.astype(np.float32)
            F_cent = kmeans_.cluster_centers_
        else:
            ########### Euclidean Clustering ##############
            kmeans_ = KMeans(n_clusters=n_clusters, n_init=500, init="k-means++").fit(F)
            kmeans_.cluster_centers_ = kmeans_.cluster_centers_.astype(np.float32)
            F_cent = kmeans_.cluster_centers_

        FCS_ = self.dFC_vec2mat(F_cent, N=n_regions)
        return FCS_, kmeans_

    def estimate_FCS(self, time_series):

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4FCS(time_series)

        # start timing
        tic = time.time()

        base_dFC = None
        if self.params["clstr_base_measure"] == "Time-Freq":
            base_dFC = TIME_FREQ(**self.params)
        if self.params["clstr_base_measure"] == "SlidingWindow":
            base_dFC = SLIDING_WINDOW(**self.params)

        # 1-level clustering
        # dFC_raw = base_dFC.estimate_dFC( \
        #         time_series=time_series \
        #         )
        # self.FCS_, self.kmeans_ = self.cluster_FC( \
        # dFC_raw.get_dFC_mat(TRs=self.dFC_raw.TR_array), \
        # n_regions = dFC_raw.n_regions \
        # )

        # 2-level clustering
        SUBJECTs = time_series.subj_id_lst
        FCS_1st_level = None
        SW_dFC = None
        for subject in SUBJECTs:

            dFC_raw = base_dFC.estimate_dFC(
                time_series=time_series.get_subj_ts(subjs_id=subject)
            )

            # test
            if dFC_raw.n_time < self.params["n_subj_clstrs"]:
                print(
                    "Number of subject-level clusters cannot be more than SW dFC samples! n_subj_clstrs was changed to "
                    + str(dFC_raw.n_time)
                    + ". This change will cause problems in similarity assessment."
                )
                self.params["n_subj_clstrs"] = dFC_raw.n_time

            FCS, _ = self.cluster_FC(
                FCS_raw=dFC_raw.get_dFC_mat(TRs=dFC_raw.TR_array),
                n_clusters=self.params["n_subj_clstrs"],
                n_regions=dFC_raw.n_regions,
            )

            if SW_dFC is None:
                SW_dFC = dFC_raw.get_dFC_mat(TRs=dFC_raw.TR_array)
            else:
                SW_dFC = np.concatenate(
                    (SW_dFC, dFC_raw.get_dFC_mat(TRs=dFC_raw.TR_array)), axis=0
                )
            if FCS_1st_level is None:
                FCS_1st_level = FCS
            else:
                FCS_1st_level = np.concatenate((FCS_1st_level, FCS), axis=0)

        self.FCS_, self.kmeans_ = self.cluster_FC(
            FCS_raw=FCS_1st_level,
            n_clusters=self.params["n_states"],
            n_regions=dFC_raw.n_regions,
        )
        self.Z = self.kmeans_.predict(self.dFC_mat2vec(SW_dFC).astype(np.float32))

        # mean activation of states
        self.set_mean_activity(time_series)

        # record time
        self.set_FCS_fit_time(time.time() - tic)

        return self

    def estimate_dFC(self, time_series):

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        assert (
            len(time_series.subj_id_lst) == 1
        ), "this function takes only one subject as input."

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        base_dFC = None
        if self.params["clstr_base_measure"] == "Time-Freq":
            base_dFC = TIME_FREQ(**self.params)
        if self.params["clstr_base_measure"] == "SlidingWindow":
            base_dFC = SLIDING_WINDOW(**self.params)

        dFC_raw = base_dFC.estimate_dFC(time_series=time_series)

        F = self.dFC_mat2vec(dFC_raw.get_dFC_mat(TRs=dFC_raw.TR_array))

        # The code below is similar for both clustering methods,
        # but is kept this way for clarity.
        if self.params["clstr_distance"] == "manhattan":
            ########### Manhattan Clustering ##############
            Z = self.kmeans_.predict(F.astype(np.float32))
        else:
            ########### Euclidean Clustering ##############
            Z = self.kmeans_.predict(F.astype(np.float32))

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(
            FCSs=self.FCS_,
            FCS_idx=Z,
            TS_info=time_series.info_dict,
            TR_array=dFC_raw.TR_array,
        )

        return dFC


################################################################################



import os
import time

import hdf5storage
import numpy as np
import scipy.io as sio
from functions.dFC_funcs import *

os.environ["MKL_NUM_THREADS"] = "64"
os.environ["NUMEXPR_NUM_THREADS"] = "64"
os.environ["OMP_NUM_THREADS"] = "64"

print(
    "################################# subject-level dFC assessment CODE started running ... #################################"
)

################################# Parameters #################################
###### DATA PARAMETERS ######

input_root = "./"
output_root = "./"

################################# LOAD #################################

dFC_analyzer = np.load(input_root + "dFC_analyzer.npy", allow_pickle="TRUE").item()
data_loader = np.load(input_root + "data_loader.npy", allow_pickle="TRUE").item()

################################# LOAD FIT MEASURES #################################

if dFC_analyzer.MEASURES_fit_lst == []:
    ALL_RECORDS = os.listdir(input_root + "fitted_MEASURES/")
    ALL_RECORDS = [i for i in ALL_RECORDS if "MEASURE" in i]
    ALL_RECORDS.sort()
    MEASURES_fit_lst = list()
    for s in ALL_RECORDS:
        fit_measure = np.load(
            input_root + "fitted_MEASURES/" + s, allow_pickle="TRUE"
        ).item()
        MEASURES_fit_lst.append(fit_measure)
    dFC_analyzer.set_MEASURES_fit_lst(MEASURES_fit_lst)
    print("fitted MEASURES loaded ...")
    # np.save('./dFC_analyzer.npy', dFC_analyzer)

################################# LOAD DATA #################################

task_id = int(os.getenv("SGE_TASK_ID"))
subj_id = data_loader.SUBJECTS[task_id - 1]  # SGE_TASK_ID starts from 1 not 0

BOLD = data_loader.load(subj_id2load=subj_id)

################################# dFC ASSESSMENT #################################

tic = time.time()
print("Measurement Started ...")

print("dFCM estimation started...")
dFCM_dict = dFC_analyzer.subj_lvl_dFC_assess(time_series_dict=BOLD)
print("dFCM estimation done.")

print(f"Measurement required {time.time() - tic:0.3f} seconds.")

################################# SAVE DATA #################################

# folder = output_root+'dFC_assessed/SUBJ_'+str(subj_id)
# if not os.path.exists(folder):
#     os.makedirs(folder)

# for dFCM_id, dFCM in enumerate(dFCM_dict['dFCM_lst']):
#     np.save(folder+'/dFCM_'+str(dFCM_id)+'.npy', dFCM)

################################# SIMILARITY MEASUREMENT #################################

similarity_assessment = SIMILARITY_ASSESSMENT(dFCM_lst=dFCM_dict["dFCM_lst"])

tic = time.time()
print("Measurement Started ...")

print("Similarity measurement started...")
SUBJ_output = similarity_assessment.run(
    FILTERS=dFC_analyzer.hyper_param_info, downsampling_method="default"
)
print("Similarity measurement done.")

print(f"Measurement required {time.time() - tic:0.3f} seconds.")

# Save
folder = output_root + "similarity_measured"
if not os.path.exists(folder):
    os.makedirs(folder)

np.save(folder + "/SUBJ_" + str(subj_id) + "_output.npy", SUBJ_output)

#################################################################################



import os
import time
import warnings

import numpy as np

from pydfc import MultiAnalysis, data_loader

warnings.simplefilter("ignore")

os.environ["MKL_NUM_THREADS"] = "16"
os.environ["NUMEXPR_NUM_THREADS"] = "16"
os.environ["OMP_NUM_THREADS"] = "16"

################################# Parameters #################################

# Data parameters
# main_root = '../../DATA/ds002785/' # for local
main_root = "../../../DATA/task-based/openneuro/ds002785/"  # for server

# subjects used for dFC assessment do not need to be the same as those used for FCS_estimate
# you can set the new roi root and data load parameters here:
roi_root = f"{main_root}/derivatives/ROI_timeseries"
fitted_measures_root = f"{main_root}/derivatives/fitted_MEASURES"
output_root = f"{main_root}/derivatives/dFC_assessed"

# for consistency we use 0 for resting state
TASKS = [
    "task-restingstate",
    "task-anticipation",
    "task-emomatching",
    "task-faces",
    "task-gstroop",
    "task-workingmemory",
]

# find all subjects across all tasks
SUBJECTS = data_loader.find_subj_list(data_root=roi_root, sessions=TASKS)

# job_id selects the subject
job_id = int(os.getenv("SGE_TASK_ID"))
if job_id > len(SUBJECTS):
    print("job_id > len(SUBJECTS)")
    exit()
subj_id = SUBJECTS[job_id - 1]  # SGE_TASK_ID starts from 1 not 0

for task in TASKS:

    MA = np.load(
        f"{fitted_measures_root}/{task}/multi_analysis.npy", allow_pickle="TRUE"
    ).item()

    # check if the subject has this task
    SUBJECTS_with_this_task = data_loader.find_subj_list(
        data_root=roi_root, sessions=[task]
    )
    if not subj_id in SUBJECTS_with_this_task:
        print(f"subject {subj_id} not in the list of subjects with task {task}")
        continue

    ################################# LOAD FIT MEASURES #################################

    ALL_RECORDS = os.listdir(f"{fitted_measures_root}/{task}/")
    ALL_RECORDS = [i for i in ALL_RECORDS if "MEASURE" in i]
    ALL_RECORDS.sort()
    MEASURES_fit_lst = list()
    for s in ALL_RECORDS:
        fit_measure = np.load(
            f"{fitted_measures_root}/{task}/{s}", allow_pickle="TRUE"
        ).item()
        MEASURES_fit_lst.append(fit_measure)
    MA.set_MEASURES_fit_lst(MEASURES_fit_lst)
    print("fitted MEASURES loaded ...")

    ################################# LOAD DATA #################################

    print(
        f"subject-level dFC assessment CODE started running ... for task {task} of subject {subj_id} ..."
    )

    BOLD = data_loader.load_TS(
        data_root=roi_root,
        file_name="time_series.npy",
        SESSIONs=[task],
        subj_id2load=subj_id,
    )

    ################################# dFC ASSESSMENT #################################

    tic = time.time()
    print("Measurement Started ...")

    print("dFC estimation started...")
    dFC_dict = MA.subj_lvl_dFC_assess(time_series_dict=BOLD)
    print("dFC estimation done.")

    print(f"Measurement required {time.time() - tic:0.3f} seconds.")

    ################################# SAVE DATA #################################

    folder = f"{output_root}/{task}/{subj_id}"
    if not os.path.exists(folder):
        os.makedirs(folder)

    for dFC_id, dFC in enumerate(dFC_dict["dFC_lst"]):
        np.save(f"{folder}/dFC_{str(dFC_id)}.npy", dFC)

#######################################################################################



"""
Implementation of functions for Multi Analysis of dFC.

Created on Dec 3 2024
@author: Mohammad Torabi
"""

from copy import deepcopy

from joblib import Parallel, delayed

from .dfc_methods import *

################################# DATA_LOADER functions ######################################


def create_measure_obj(MEASURES_name_lst, **params):

    MEASURES_lst = list()
    for MEASURES_name in MEASURES_name_lst:

        ###### CAP ######
        if MEASURES_name == "CAP":
            measure = CAP(**params)

        ###### CONTINUOUS HMM ######
        if MEASURES_name == "ContinuousHMM":
            measure = HMM_CONT(**params)

        ###### WINDOW_LESS ######
        if MEASURES_name == "Windowless":
            measure = WINDOWLESS(**params)

        ###### SLIDING WINDOW ######
        if MEASURES_name == "SlidingWindow":
            measure = SLIDING_WINDOW(**params)

        ###### TIME FREQUENCY ######
        if MEASURES_name == "Time-Freq":
            measure = TIME_FREQ(**params)

        ###### SLIDING WINDOW + CLUSTERING ######
        if MEASURES_name == "Clustering":
            measure = SLIDING_WINDOW_CLUSTR(**params)

        ###### DISCRETE HMM ######
        if MEASURES_name == "DiscreteHMM":
            measure = HMM_DISC(**params)

        MEASURES_lst.append(measure)

    return MEASURES_lst


def measures_initializer(MEASURES_name_lst, params_methods, alter_hparams):
    """
    - this will test values in alter_hparams other than
        values already in params_methods. values in params_methods
        will be considered the reference
    sample:
    hyper_params = { \
        'n_states': [6, 12, 16], \
        'normalization': [True], \
        'num_subj': [50, 100, 395], \
        'num_select_nodes': [50, 100, 333], \
        'num_time_point': [500, 800, 1200], \
        'Fs_ratio': [0.50, 1.00, 1.50], \
        'noise_ratio': [0.00, 0.50, 1.00], \
        'num_realization': [1, 2, 3], \
        }

        MEASURES_name_lst = ( \
            'SlidingWindow', \
            'Time-Freq', \
            'CAP', \
            'ContinuousHMM', \
            'Windowless', \
            'Clustering', \
            'DiscreteHMM' \
            )
    """

    # a list of MEASURES with default parameter values
    MEASURES_lst = create_measure_obj(
        MEASURES_name_lst=MEASURES_name_lst, **params_methods
    )

    # adding MEASURES with alternative parameter values
    hyper_param_info = {}
    hyper_param_info["default_values"] = params_methods
    for hyper_param in alter_hparams:
        for value in alter_hparams[hyper_param]:
            params = deepcopy(params_methods)
            params[hyper_param] = value
            hyper_param_info[hyper_param + "_" + str(value)] = deepcopy(params)
            new_MEASURES = create_measure_obj(
                MEASURES_name_lst=MEASURES_name_lst, **params
            )
            for new_measure in new_MEASURES:
                flag = 0
                for MEASURE in MEASURES_lst:
                    if new_measure.issame(MEASURE):
                        flag = 1
                if flag == 0:
                    MEASURES_lst.append(new_measure)

    return MEASURES_lst, hyper_param_info


def get_SB_MEASURES_lst(MEASURES_lst):
    """returns state_based measures"""
    SB_MEASURES = list()
    for measure in MEASURES_lst:
        if measure.is_state_based:
            SB_MEASURES.append(measure)
    return SB_MEASURES


def get_DD_MEASURES_lst(MEASURES_lst):
    """returns data_driven measures"""
    DD_MEASURES = list()
    for measure in MEASURES_lst:
        if not measure.is_state_based:
            DD_MEASURES.append(measure)
    return DD_MEASURES


def estimate_group_FCS(time_series, MEASURES_lst, n_jobs=None, verbose=0, backend="loky"):

    if n_jobs is None:
        MEASURES_fit_lst = list()
        for measure in MEASURES_lst:
            MEASURES_fit_lst.append(measure.estimate_FCS(time_series=time_series))
    else:
        MEASURES_fit_lst = Parallel(
            n_jobs=n_jobs,
            verbose=verbose,
            backend=backend,
        )(
            delayed(measure.estimate_FCS)(time_series=time_series)
            for measure in MEASURES_lst
        )

    return MEASURES_fit_lst


##################### dFC ASSESSMENT ######################


def group_dFC_assess(
    time_series, MEASURES_fit_lst, n_jobs=None, verbose=0, backend="loky"
):
    """
    assess dFC for all subjects using all measures
    and a single time_series
    """
    SUBJECTs = time_series.subj_id_lst

    OUT = list()
    for subject in SUBJECTs:
        OUT.append(
            subj_lvl_dFC_assess(
                time_series=time_series.get_subj_ts(subjs_id=subject),
                MEASURES_fit_lst=MEASURES_fit_lst,
                n_jobs=n_jobs,
                verbose=verbose,
                backend=backend,
            )
        )

    return OUT


def subj_lvl_dFC_assess(
    time_series, MEASURES_fit_lst, n_jobs=None, verbose=0, backend="loky"
):
    """
    assess dFC for a single subject using all measures and a
    single time_series
    """

    dFC_dict = {}

    if n_jobs is None:
        dFC_lst = list()
        for measure in MEASURES_fit_lst:
            dFC_lst.append(measure.estimate_dFC(time_series=time_series))
    else:
        dFC_lst = Parallel(
            n_jobs=n_jobs,
            verbose=verbose,
            backend=backend,
        )(
            delayed(measure.estimate_dFC)(time_series=time_series)
            for measure in MEASURES_fit_lst
        )

    dFC_dict["dFC_lst"] = dFC_lst

    return dFC_dict


##############################################################################################################



import numpy as np
import pytest

from pydfc import TIME_SERIES


def test_create_time_series():
    # create a random data with 100 regions and 1000 time points
    data = np.random.rand(100, 1000)
    # create a random locs with 100 regions and 3 coordinates
    locs = np.random.rand(100, 3)
    # create a random node_labels list with 100 regions
    node_labels = [f"Region {i}" for i in range(100)]
    time_series = TIME_SERIES(
        data=data,
        subj_id="sub-0001",
        Fs=2,
        locs=locs,
        node_labels=node_labels,
    )

    assert time_series.data.shape == (100, 1000)
    assert time_series.subj_id_lst == ["sub-0001"]
    assert time_series.Fs == 2
    assert time_series.n_time == 1000
    assert time_series.n_regions == 100
    assert np.all(time_series.nodes_lst == np.arange(0, time_series.n_regions, dtype=int))
    assert np.all(time_series.time == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2))
    assert time_series.data_dict.keys() == {"sub-0001"}
    assert np.all(time_series.data_dict["sub-0001"]["data"] == data)


def test_append_ts():
    # create a random data with 100 regions and 1000 time points
    data_1 = np.random.rand(100, 1000)
    # create a random locs with 100 regions and 3 coordinates
    locs = np.random.rand(100, 3)
    # create a random node_labels list with 100 regions
    node_labels = [f"Region {i}" for i in range(100)]
    time_series = TIME_SERIES(
        data=data_1,
        subj_id="sub-0001",
        Fs=2,
        locs=locs,
        node_labels=node_labels,
    )

    # create a random data with 100 regions and 1000 time points
    data_2 = np.random.rand(100, 1000)
    time_series.append_ts(
        new_time_series=data_2,
        subj_id="sub-0002",
    )

    assert time_series.data.shape == (100, 2000)
    assert time_series.subj_id_lst == ["sub-0001", "sub-0002"]
    assert time_series.Fs == 2
    assert time_series.n_time is None
    assert time_series.n_regions == 100
    assert np.all(time_series.nodes_lst == np.arange(0, time_series.n_regions, dtype=int))
    assert time_series.time is None
    assert time_series.data_dict.keys() == {"sub-0001", "sub-0002"}
    assert np.all(time_series.data_dict["sub-0001"]["data"] == data_1)
    assert np.all(time_series.data_dict["sub-0002"]["data"] == data_2)
    assert np.all(
        time_series.data_dict["sub-0001"]["time_array"]
        == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2)
    )
    assert np.all(
        time_series.data_dict["sub-0002"]["time_array"]
        == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2)
    )

    # check if visualization will raise warning correctly
    with pytest.warns(
        UserWarning, match="Multiple subjects are not supported in visualization."
    ):
        time_series.visualize()


def test_concat_ts():
    # create a random data with 100 regions and 1000 time points
    data_1 = np.random.rand(100, 1000)
    # create a random locs with 100 regions and 3 coordinates
    locs = np.random.rand(100, 3)
    # create a random node_labels list with 100 regions
    node_labels = [f"Region {i}" for i in range(100)]
    time_series = TIME_SERIES(
        data=data_1,
        subj_id="sub-0001",
        Fs=2,
        locs=locs,
        node_labels=node_labels,
    )

    # create a random data with 100 regions and 1000 time points
    data_2 = np.random.rand(100, 1000)
    locs_2 = np.random.rand(100, 3)
    node_labels_2 = [f"Parcel {i}" for i in range(100)]

    time_series_2 = TIME_SERIES(
        data=data_2,
        subj_id="sub-0002",
        Fs=1,
        locs=locs,
        node_labels=node_labels,
    )

    # check Fs mismatch assert error
    with pytest.raises(AssertionError, match="Fs mismatch!"):
        time_series.concat_ts(time_series_2)

    time_series_2 = TIME_SERIES(
        data=data_2,
        subj_id="sub-0002",
        Fs=2,
        locs=locs_2,
        node_labels=node_labels,
    )
    # check locs mismatch assert error
    with pytest.raises(AssertionError, match="locs mismatch!"):
        time_series.concat_ts(time_series_2)

    time_series_2 = TIME_SERIES(
        data=data_2,
        subj_id="sub-0002",
        Fs=2,
        locs=locs,
        node_labels=node_labels_2,
    )
    # check node_labels mismatch assert error
    with pytest.raises(AssertionError, match="node_labels mismatch!"):
        time_series.concat_ts(time_series_2)

    time_series_2 = TIME_SERIES(
        data=data_2,
        subj_id="sub-0002",
        Fs=2,
        locs=locs,
        node_labels=node_labels,
    )
    time_series.concat_ts(time_series_2)

    assert time_series.data.shape == (100, 2000)
    assert time_series.subj_id_lst == ["sub-0001", "sub-0002"]
    assert time_series.Fs == 2
    assert time_series.n_time is None
    assert time_series.n_regions == 100
    assert np.all(time_series.nodes_lst == np.arange(0, 100, dtype=int))
    assert time_series.time is None
    assert time_series.data_dict.keys() == {"sub-0001", "sub-0002"}
    assert np.all(time_series.data_dict["sub-0001"]["data"] == data_1)
    assert np.all(time_series.data_dict["sub-0002"]["data"] == data_2)
    assert np.all(
        time_series.data_dict["sub-0001"]["time_array"]
        == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2)
    )
    assert np.all(
        time_series.data_dict["sub-0002"]["time_array"]
        == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2)
    )


def test_get_subj_ts():
    # create a random data with 100 regions and 1000 time points
    data_1 = np.random.rand(100, 1000)
    # create a random locs with 100 regions and 3 coordinates
    locs = np.random.rand(100, 3)
    # create a random node_labels list with 100 regions
    node_labels = [f"Region {i}" for i in range(100)]
    time_series = TIME_SERIES(
        data=data_1,
        subj_id="sub-0001",
        Fs=2,
        locs=locs,
        node_labels=node_labels,
    )
    # create a random data with 100 regions and 1000 time points
    data_2 = np.random.rand(100, 1000)
    time_series_2 = TIME_SERIES(
        data=data_2,
        subj_id="sub-0002",
        Fs=2,
        locs=locs,
        node_labels=node_labels,
    )
    time_series.concat_ts(time_series_2)

    subj_ts = time_series.get_subj_ts(subjs_id="sub-0001")
    assert subj_ts.data.shape == (100, 1000)
    assert subj_ts.subj_id_lst == ["sub-0001"]
    assert subj_ts.Fs == 2
    assert subj_ts.n_time == 1000
    assert subj_ts.n_regions == 100
    assert np.all(subj_ts.nodes_lst == np.arange(0, 100, dtype=int))
    assert np.all(subj_ts.time == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2))
    assert subj_ts.data_dict.keys() == {"sub-0001"}
    assert np.all(subj_ts.data_dict["sub-0001"]["data"] == data_1)

    subj_ts = time_series.get_subj_ts(subjs_id="sub-0002")
    assert subj_ts.data.shape == (100, 1000)
    assert subj_ts.subj_id_lst == ["sub-0002"]
    assert subj_ts.Fs == 2
    assert subj_ts.n_time == 1000
    assert subj_ts.n_regions == 100
    assert np.all(subj_ts.nodes_lst == np.arange(0, 100, dtype=int))
    assert np.all(subj_ts.time == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2))
    assert subj_ts.data_dict.keys() == {"sub-0002"}
    assert np.all(subj_ts.data_dict["sub-0002"]["data"] == data_2)

    # check that the original time_series is not changed
    assert time_series.data.shape == (100, 2000)
    assert time_series.subj_id_lst == ["sub-0001", "sub-0002"]
    assert time_series.Fs == 2
    assert time_series.n_time is None
    assert time_series.n_regions == 100
    assert np.all(time_series.nodes_lst == np.arange(0, 100, dtype=int))
    assert time_series.time is None
    assert time_series.data_dict.keys() == {"sub-0001", "sub-0002"}
    assert np.all(time_series.data_dict["sub-0001"]["data"] == data_1)
    assert np.all(time_series.data_dict["sub-0002"]["data"] == data_2)
    assert np.all(
        time_series.data_dict["sub-0001"]["time_array"]
        == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2)
    )
    assert np.all(
        time_series.data_dict["sub-0002"]["time_array"]
        == 1 / 2 + np.arange(0, 1000 / 2, 1 / 2)
    )


def test_select_nodes():
    # create a random data with 100 regions and 1000 time points
    data = np.random.rand(100, 1000)
    # create a random locs with 100 regions and 3 coordinates
    locs = np.random.rand(100, 3)
    # create a random node_labels list with 100 regions
    node_labels = [f"Region {i}" for i in range(100)]
    time_series = TIME_SERIES(
        data=data,
        subj_id="sub-0001",
        Fs=2,
        locs=locs,
        node_labels=node_labels,
    )

    # select 5 nodes
    nodes_idx = np.array([0, 1, 23, 43, 87])
    time_series.select_nodes(nodes_idx=nodes_idx)
    assert time_series.data.shape == (5, 1000)
    assert np.all(time_series.data == data[nodes_idx, :])
    assert time_series.n_regions == 5
    assert np.all(time_series.nodes_lst == nodes_idx)
    assert np.all(time_series.locs == locs[nodes_idx])
    assert np.all(
        time_series.node_labels
        == ["Region 0", "Region 1", "Region 23", "Region 43", "Region 87"]
    )
    assert np.all(time_series.nodes_selection_ == nodes_idx)



import os
import time

import hdf5storage
import numpy as np
import scipy.io as sio
from functions.dFC_funcs import *

os.environ["MKL_NUM_THREADS"] = "64"
os.environ["NUMEXPR_NUM_THREADS"] = "64"
os.environ["OMP_NUM_THREADS"] = "64"

################################# Parameters #################################

###### DATA PARAMETERS ######
DATA_type = "Gordon"  # 'Gordon' or 'simulated' or 'ICA'

output_root = "./../../../../../RESULTs/methods_implementation/"
# output_root = '/data/origami/dFC/RESULTs/methods_implementation/'
# output_root = '/Users/mte/Documents/McGill/Project/dFC/RESULTs/methods_implementation/'

data_root_simul = "./../../../../DATA/TVB data/"
data_root_sample = "./sampleDATA/"
data_root_gordon = "./../../../../DATA/HCP/HCP_Gordon/"
data_root_ica = (
    "./../../../../DATA/HCP/HCP_PTN1200/node_timeseries/3T_HCP1200_MSMAll_d50_ts2/"
)

num_subj = 2
select_nodes = True
rand_node_slct = False
num_select_nodes = 50

###### MEASUREMENT PARAMETERS ######

# W is in sec

params_methods = {  # Sliding Parameters
    "W": 44,
    "n_overlap": 0.5,  # State Parameters
    "n_states": 6,
    "n_subj_clstrs": 20,
    "n_hid_states": 4,  # Parallelization Parameters
    "n_jobs": 2,
    "verbose": 0,
    "backend": "loky",
}

###### SIMILARITY PARAMETERS ######

sim_assess_params = {
    "run_analysis": True,
    "num_samples": 100,
    "matching_method": "score",
    "n_jobs": 8,
    "backend": "loky",
}

###### SIMILARITY PARAMETERS ######

dyn_conn_det_params = {
    "run_analysis": False,
    "N": 30,
    "L": 1200,
    "p": 100,
    "n_jobs": 8,
    "backend": "loky",
}

###### dFC ANALYZER PARAMETERS ######

params_dFC_analyzer = {  # VISUALIZATION
    "vis_TR_idx": list(range(10, 20, 1)),
    "save_image": True,
    "output_root": output_root,  # Parallelization Parameters
    "n_jobs": 8,
    "verbose": 0,
    "backend": "loky",  # Similarity Assessment Parameters
    "sim_assess_params": sim_assess_params,  # Dynamic Connection Detector Parameters
    "dyn_conn_det_params": dyn_conn_det_params,
}


################################# FINDING SUBJECTs LIST #################################

# ICA
ALL_RECORDS = os.listdir(data_root_ica)
ALL_RECORDS = [i for i in ALL_RECORDS if ".txt" in i]
ALL_RECORDS.sort()
SUBJECTS_ica = list()
for s in ALL_RECORDS:
    num = s[: s.find(".")]
    SUBJECTS_ica.append(num)
SUBJECTS_ica = list(set(SUBJECTS_ica))
SUBJECTS_ica.sort()

# GORDON
ALL_RECORDS = os.listdir(data_root_gordon)
ALL_RECORDS = [i for i in ALL_RECORDS if "Rest" in i]
ALL_RECORDS.sort()
SUBJECTS_gordon = list()
for s in ALL_RECORDS:
    num = s[: s.find("_")]
    SUBJECTS_gordon.append(num)
SUBJECTS_gordon = list(set(SUBJECTS_gordon))
SUBJECTS_gordon.sort()

SUBJECTS = intersection(SUBJECTS_gordon, SUBJECTS_ica)

print(
    str(len(SUBJECTS))
    + " subjects were found. "
    + str(num_subj)
    + " subjects were selected."
)

SUBJECTS = SUBJECTS[0:num_subj]

################################# Load ICA BOLD data (HCP) #################################

if DATA_type == "ICA":

    SESSIONs = ["session_1"]
    BOLD_Fs = 1 / 0.72

    BOLD = {}
    for session in SESSIONs:
        BOLD[session] = None
        for subject in SUBJECTS:
            time_series = np.loadtxt(data_root_ica + subject + ".txt", dtype="float64")
            time_series = time_series.T

            # time_series = time_series - np.repeat(np.mean(time_series, axis=1)[:,None], time_series.shape[1], axis=1) # ???????????????????????

            if BOLD[session] is None:
                BOLD[session] = TIME_SERIES(
                    data=time_series,
                    subj_id=subject,
                    Fs=BOLD_Fs,
                    TS_name="BOLD ICA",
                    session_name=session,
                )
            else:
                BOLD[session].append_ts(new_time_series=time_series, subj_id=subject)

        print(BOLD[session].n_regions, BOLD[session].n_time)

################################# Load Real BOLD data (HCP) #################################

if DATA_type == "Gordon":

    SESSIONs = ["Rest1_LR", "Rest1_RL", "Rest2_LR", "Rest2_RL"]
    BOLD_Fs = 1 / 0.72

    # LOAD Region Location DATA

    locs = sio.loadmat(data_root_gordon + "Gordon333_LOCS.mat")
    locs = locs["locs"]

    # LOAD Region Data

    file = data_root_gordon + "Gordon333_Key.txt"
    f = open(file, "r")

    atlas_data = []
    for line in f:
        row = line.split()
        atlas_data.append(row)

    BOLD = {}
    for session in SESSIONs:
        BOLD[session] = None
        for subject in SUBJECTS:

            subj_fldr = subject + "_" + session

            # LOAD BOLD Data

            DATA = hdf5storage.loadmat(
                data_root_gordon + subj_fldr + "/ROI_data_Gordon_333_surf.mat"
            )
            time_series = DATA["ROI_data"]

            time_series = time_series.T

            # time_series = time_series - np.repeat(np.mean(time_series, axis=1)[:,None], time_series.shape[1], axis=1) # ???????????????????????

            for n in range(time_series.shape[0]):
                time_series[n, :] = time_series[n, :] - np.mean(time_series[n, :])
                time_series[n, :] = np.divide(
                    time_series[n, :], np.std(time_series[n, :])
                )

            if BOLD[session] is None:
                BOLD[session] = TIME_SERIES(
                    data=time_series,
                    subj_id=subject,
                    Fs=BOLD_Fs,
                    locs=locs,
                    nodes_info=atlas_data,
                    TS_name="BOLD Real",
                    session_name=session,
                )
            else:
                BOLD[session].append_ts(new_time_series=time_series, subj_id=subject)

        print("*** Session " + session + ": ")
        print(
            "number of regions= "
            + str(BOLD[session].n_regions)
            + ", number of TRs= "
            + str(BOLD[session].n_time)
        )

        # select nodes

        if select_nodes:
            if rand_node_slct:
                nodes_idx = np.random.choice(
                    range(BOLD[session].n_regions), size=num_select_nodes, replace=False
                )
                nodes_idx.sort()
            else:
                nodes_idx = np.array(list(range(47, 88)) + list(range(224, 263)))
            BOLD[session].select_nodes(nodes_idx=nodes_idx)

            print("number of regions selected= " + str(BOLD[session].n_regions))

################################# Load Sample BOLD data #################################

if DATA_type == "sample":

    BOLD_Fs = 1 / 0.5

    ###### BOLD DATA ######
    time_BOLD = np.load(data_root_sample + "bold_time.npy") / 1e3
    time_series = np.load(data_root_sample + "bold_data.npy")

    time_series = time_series.T

    BOLD = None
    for subject in range(5):

        if BOLD is None:
            BOLD = TIME_SERIES(
                data=time_series[:, (subject) * 1200 : (subject + 1) * 1200],
                subj_id=str(subject + 1),
                Fs=BOLD_Fs,
                time_array=time_BOLD[(subject) * 1200 : (subject + 1) * 1200],
                TS_name="BOLD Sample",
            )
        else:
            BOLD.append_ts(
                new_time_series=time_series[:, (subject) * 1200 : (subject + 1) * 1200],
                time_array=time_BOLD[(subject) * 1200 : (subject + 1) * 1200],
                subj_id=str(subject + 1),
            )

    print(BOLD.n_regions, BOLD.n_time)


################################# Load Simulated BOLD data #################################

if DATA_type == "simulated":
    time_BOLD = np.load(data_root_simul + "bold_time.npy") / 1e3
    time_series_BOLD = np.load(data_root_simul + "bold_data.npy")

    BOLD = TIME_SERIES(
        data=time_series_BOLD.T,
        subj_id="1",
        Fs=BOLD_Fs,
        time_array=time_BOLD,
        TS_name="BOLD Simulation",
    )

################################# Load Simulated Tavg data #################################

if DATA_type == "simulated":
    time_Tavg = np.load(data_root_simul + "TVB data/tavg_time.npy") / 1e3
    time_series_Tavg = np.load(data_root_simul + "TVB data/tavg_data.npy")

    TAVG = TIME_SERIES(
        data=time_series_Tavg.T,
        subj_id="1",
        Fs=200,
        time_array=time_Tavg,
        TS_name="Tavg Simulation",
    )

################################# Visualize BOLD #################################

for session in BOLD:
    BOLD[session].visualize(
        start_time=0,
        end_time=50,
        nodes_lst=list(range(10)),
        save_image=params_dFC_analyzer["save_image"],
        output_root=output_root + "BOLD_signal_" + session,
    )

################################# Measure dFC #################################

###### CONTINUOUS HMM ######
hmm_cont = HMM_CONT(**params_methods)

###### WINDOW_LESS ######
windowless = WINDOWLESS(**params_methods)

###### SLIDING WINDOW ######
sw_pc = SLIDING_WINDOW(sw_method="pear_corr", **params_methods)
sw_mi = SLIDING_WINDOW(sw_method="MI", **params_methods)

###### TIME FREQUENCY ######
time_freq_cwt = TIME_FREQ(method="CWT_mag", **params_methods)
time_freq_wtc = TIME_FREQ(method="WTC", **params_methods)

###### SLIDING WINDOW + CLUSTERING ######
swc_pc = SLIDING_WINDOW_CLUSTR(base_method="pear_corr", **params_methods)

###### DISCRETE HMM ######
hmm_disc_pc = HMM_DISC(base_method="pear_corr", **params_methods)


MEASURES = [
    hmm_cont,
    windowless,
    sw_pc,  # sw_mi, \
    # sw_gLasso, \
    # time_freq_cwt, \
    # time_freq_cwt_r, \
    # time_freq_wtc, \
    swc_pc,  # swc_gLasso, \
    hmm_disc_pc,  # hmm_disc_gLasso, \
]

dFC_analyzer = DFC_ANALYZER(
    MEASURES_lst=MEASURES,
    analysis_name="reproducibility assessment",
    **params_dFC_analyzer,
)

tic = time.time()
print("Measurement Started ...")
SUBJs_dFC_session_sim_dict = dFC_analyzer.analyze(time_series_dict=BOLD)

# Save
np.save("./dFC_analyzer.npy", dFC_analyzer)
np.save("./dFC_session_sim.npy", SUBJs_dFC_session_sim_dict)

print(f"Measurement required {time.time() - tic:0.3f} seconds.")

################################# SIMILARITY ASSESSMENT #################################

# print_dict(dFC_analyzer.methods_corr)
dFC_analyzer.similarity_analyze(SUBJs_dFC_session_sim_dict)

################################# STATE MATCH #################################

state_match = dFC_analyzer.state_match()
# Save
np.save("./state_match.npy", state_match)

################################# POST ANALYSIS #################################

dFC_analyzer.post_analyze()

################################# SAMPLE CHECK #################################

session = "Rest1_LR"

measure_pairs = [
    ["ContinuousHMM", "Windowless"],
    ["ContinuousHMM", "Clustering_pear_corr"],
    ["ContinuousHMM", "DiscreteHMM_pear_corr"],
]
subj_id = np.unique(BOLD[session].subj_id_array)[1]
time_series = get_subj_ts_dict(BOLD, subj_id=subj_id)[session]

# if params_dFC_analyzer['n_jobs'] is None:
#     dFCM_lst = list()
#     for measure in dFC_analyzer.MEASURES_fit_lst[session]:
#         dFCM_lst.append( \
#             measure.estimate_dFCM(time_series=time_series) \
#         )
# else:
#     dFCM_lst = Parallel( \
#         n_jobs=params_dFC_analyzer['n_jobs'], verbose=params_dFC_analyzer['verbose'], backend=params_dFC_analyzer['backend'])( \
#         delayed(measure.estimate_dFCM)(time_series=time_series) \
#             for measure in dFC_analyzer.MEASURES_fit_lst[session])

for measure_pair in measure_pairs:
    measure_1 = dFC_analyzer.MEASURES_fit_dict[session][measure_pair[0]]
    measure_2 = dFC_analyzer.MEASURES_fit_dict[session][measure_pair[1]]

    # re-run estimate_dFCM for a sample subject

    print("dFCM estimation started...")
    dFCM_i = measure_1.estimate_dFCM(time_series=time_series)
    print("dFCM estimation done.")

    print("dFCM estimation started...")
    dFCM_j = measure_2.estimate_dFCM(time_series=time_series)
    print("dFCM estimation done.")

    # TRs = TR_intersection(dFCM_lst)
    TRs = TR_intersection([dFCM_i, dFCM_j])

    # state match visualization on the sample subject with the state matching result from all subjects
    dFC_analyzer.state_transition_analyze(
        dFCM_i=dFCM_i,
        dFCM_j=dFCM_j,
        state_match_dict=state_match[session]["method_pairs"][measure_1.measure_name][
            measure_2.measure_name
        ],
        TRs=TRs,
        subject=subj_id,
        session=session,
        matching_method="score",
        verb=True,
    )  # matching_method = score / FCS / transition

#########################################################################################



"""The :mod:`pydfc.dfc_methods` contains dFC methods objects."""

from .base_dfc_method import BaseDFCMethod
from .cap import CAP
from .continuous_hmm import HMM_CONT
from .discrete_hmm import HMM_DISC
from .sliding_window import SLIDING_WINDOW
from .sliding_window_clustr import SLIDING_WINDOW_CLUSTR
from .time_freq import TIME_FREQ
from .windowless import WINDOWLESS

__all__ = [
    "BaseDFCMethod",
    "CAP",
    "SLIDING_WINDOW_CLUSTR",
    "HMM_CONT",
    "HMM_DISC",
    "SLIDING_WINDOW",
    "TIME_FREQ",
    "WINDOWLESS",
]



import os
from copy import deepcopy
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

# sys.path.append('./BIC_codes/')
from functions.dFC_funcs import *
from functions.post_analysis_funcs import *

print(" VISUALIZATION IN PROGRESS ... ")

################################# LOAD RESULTS #################################

output_root = "./../../../../RESULTs/methods_implementation/server/methods_implementation/final_figures/"
# output_root = './output/'
save_image = True

# the dictionary that collects all RESULTS
ALL_RESULTS = np.load(output_root + "ALL_RESULTS.npy", allow_pickle="True").item()

################################ common variables #################################

node_networks = ALL_RESULTS["node_networks"]

cluster_colors_dict = {
    "forestgreen": ["SlidingWindow", "Time-Freq"],
    "coral": ["Clustering", "ContinuousHMM", "DiscreteHMM"],
    "crimson": ["CAP", "Windowless"],
}

extra_colors = ["khaki", "orchid", "aquamarine"]

################################# dFC SAMPLES #################################

RESULTS = ALL_RESULTS["dFC_sample"]

plot_sample_dFC(
    D=RESULTS,
    x="samples",
    title="dFC_samples",
    cmap="seismic",
    normalize=False,
    disp_diag=False,
    save_image=save_image,
    output_root=output_root + "dFC_sample/",
    fix_lim=False,
    center_0=True,
    node_networks=node_networks,
    segmented=False,
)

plot_sample_dFC(
    D=RESULTS,
    x="samples_ranked",
    title="dFC_samples_ranked",
    cmap="plasma",
    normalize=False,
    disp_diag=False,
    save_image=save_image,
    output_root=output_root + "dFC_sample/",
    fix_lim=False,
    center_0=False,
    node_networks=node_networks,
    segmented=False,
)

################################# FCS visualization #################################

for measure in ALL_RESULTS["measure_lst"]:

    visualize_FCS(
        measure,
        normalize=True,
        fix_lim=False,
        save_image=save_image,
        output_root=output_root + "FCS/",
    )

################################# RSNs visualization #################################

measure = ALL_RESULTS["measure_lst"][0]

# find ROI count in each RSN
RSNs = np.unique(node_networks)
num_ROIs = {}
for RSN in RSNs:
    num_ROIs[RSN] = np.sum([1 for node in node_networks if node == RSN])

# write to a txt file
folder = f"{output_root}RSNs/"
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder + "/ROI_count_in_RSNs.txt")
filename.touch(exist_ok=True)
text_file = open(filename, "wt")
for RSN in num_ROIs:
    text_file.write(RSN + " : " + str(num_ROIs[RSN]) + "\n")
text_file.close()

plot_rois(
    node_networks,
    measure.TS_info["nodes_locs"],
    save_image=save_image,
    output_root=f"{output_root}RSNs/",
)

################################# dFC values distributions #################################
dFC_dist_plot = True

if dFC_dist_plot:

    RESULTS = ALL_RESULTS["dFC_values_dist"]

    ############ VISUALIZE ############

    joint_dist_plot(
        data=RESULTS,
        title="dFC values distributions",
        save_image=save_image,
        output_root=output_root + "indiv_prop/",
    )

################################# dFC Similarity #################################

RESULTS = ALL_RESULTS["dFC_similarity_overall"]["spearman"]["session_Rest1_LR"][
    "overall_stat"
]

folder = output_root + "dFC_similarity"
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder + "/overall_stat.txt")
filename.touch(exist_ok=True)
text_file = open(filename, "wt")
text_file.write("Average similarity of all pairs (E_E) = " + str(RESULTS["E_E"]) + "\n")
text_file.write(
    "Variance of average similarity across pairs (VAR_E) = "
    + str(RESULTS["VAR_E"])
    + " and std = "
    + str(np.sqrt(RESULTS["VAR_E"]))
    + "\n"
)
text_file.write(
    "Average variance of similarity across subjects (E_VAR) = "
    + str(RESULTS["E_VAR"])
    + "\n"
)
text_file.write(
    "ratio of variances (VAR_E / E_VAR) = "
    + str(RESULTS["VAR_E"] / RESULTS["E_VAR"])
    + "\n"
)
text_file.close()

################# whole subject #################
"""
    - corr((all dFC timepoints of one Subj using method_i), (all dFC timepoints of one Subj using method_j)) -> avg(corr) and var(corr)
    - spearman correlation, pearson correlation, Mutual Information (MI), Euclidean Distance
    - dendogram based on avg(corr)
"""

for metric in ALL_RESULTS["dFC_similarity_overall"]:

    RESULTS = ALL_RESULTS["dFC_similarity_overall"][metric]
    filters_lst = [filter for filter in RESULTS]
    keys_lst = [key for key in RESULTS[filters_lst[0]]]

    ############ VISUALIZE ############
    label_dict = {
        "session_Rest1_LR": "session Rest1 LR",
        "session_Rest1_RL": "session Rest1 RL",
        "session_Rest2_LR": "session Rest2 LR",
        "session_Rest2_RL": "session Rest2 RL",
    }
    for key in keys_lst:
        if key == "name_lst" or key == "sim_distribution" or key == "overall_stat":
            continue
        visualize_sim_mat(
            RESULTS,
            mat_key=key,
            title=metric + " " + key,
            name_lst_key="name_lst",
            label_dict=label_dict,
            cmap="viridis",
            save_image=save_image,
            output_root=output_root + "dFC_similarity/" + metric + "/",
        )

    #### similarity values distribution ####
    label_dict = {"sim": "similarity"}
    for filter in ["session_Rest1_LR"]:
        pairwise_cat_plots(
            RESULTS[filter]["sim_distribution"],
            y="sim",
            title=metric + " total similarity distributions " + filter,
            label_dict=label_dict,
            save_image=save_image,
            output_root=output_root + "dFC_similarity/" + metric + "/",
        )

    ############ Hierarchical Clustering ############
    for filter in ["session_Rest1_LR"]:
        dist_mat = corr2distance(RESULTS[filter]["avg_mat"], metric=metric)
        Z = distance2Z(dist_mat, method="ward")

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(
            Z=Z,
            labels=RESULTS[filter]["name_lst"],
            cluster_colors_dict=cluster_colors_dict,
            extra_colors=extra_colors,
        )

        dist_mat_dendo(
            Z=Z,
            labels=RESULTS[filter]["name_lst"],
            link_colors=link_cols,
            plot_threshold=True,
            title="Hierarchical Clustering of Methods " + filter + " using " + metric,
            save_image=save_image,
            output_root=output_root + "dFC_similarity/" + metric + "/",
        )

################# Hierarchical Clstr with Confidence Interval #################
"""
    - plot the most frequent hierclstr structures across subjects
"""
min_freq = 10
Z_lst = ALL_RESULTS["Hierclstr_CI"]
measures_lst = [measure.measure_name for measure in ALL_RESULTS["measure_lst"]]

# cluster Zs to find common structures
Z_clstrs = cluster_Z(Z_lst, num_leaf=len(measures_lst))

num_clstrs = len([key for key in Z_clstrs])
num_clstrs_included = 0
for key in Z_clstrs:
    Z_clstrs[key]["distance_lst"] = np.array(Z_clstrs[key]["distance_lst"])
    if Z_clstrs[key]["freq"] > min_freq:
        num_clstrs_included += 1
        n = Z_clstrs[key]["distance_lst"].shape[0]
        avg_distances = np.mean(Z_clstrs[key]["distance_lst"], axis=0)
        std_distances = np.std(Z_clstrs[key]["distance_lst"], axis=0)

        # reconstruct avg Z
        Z = list()
        for i, tree in enumerate(Z_clstrs[key]["Z"]):
            Z.append([tree[0], tree[1], avg_distances[i], tree[3]])
        Z = np.array(Z)

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(
            Z=Z,
            labels=measures_lst,
            cluster_colors_dict=cluster_colors_dict,
            extra_colors=extra_colors,
        )

        dist_mat_dendo(
            Z,
            labels=measures_lst,
            link_colors=link_cols,
            plot_threshold=True,
            distances_CI=std_distances,
            title="Hierclstr clstr " + str(key) + "_" + str(n) + "subjects",
            save_image=save_image,
            output_root=output_root + "hierclstr_CI/",
        )

# write to a txt file
folder = output_root + "hierclstr_CI"
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder + "/hierclstr.txt")
filename.touch(exist_ok=True)
text_file = open(filename, "wt")
text_file.write("Number of structures found across subjects = " + str(num_clstrs) + "\n")
text_file.write("The min number of subjects to be included = " + str(min_freq) + "\n")
text_file.write("Number of structures included = " + str(num_clstrs_included) + "\n")
text_file.close()

################# session ANOVA #################

if "session_ANOVA" in ALL_RESULTS:

    RESULTS = ALL_RESULTS["session_ANOVA"]
    measure_lst = [measure.measure_name for measure in ALL_RESULTS["measure_lst"]]

    data = {
        "day p-values": {
            "p_values": np.nan * np.ones((len(measure_lst), len(measure_lst))),
            "measure_lst": measure_lst,
        },
        "direction p-values": {
            "p_values": np.nan * np.ones((len(measure_lst), len(measure_lst))),
            "measure_lst": measure_lst,
        },
    }

    for i, measure_i in enumerate(RESULTS):
        for j, measure_j in enumerate(RESULTS[measure_i]):
            result = RESULTS[measure_i][measure_j]

            index_i = measure_lst.index(unzip_name(measure_i))
            index_j = measure_lst.index(unzip_name(measure_j))

            data["day p-values"]["p_values"][index_i, index_j] = result["PR(>F)"][0]
            data["day p-values"]["p_values"][index_j, index_i] = result["PR(>F)"][0]
            data["direction p-values"]["p_values"][index_i, index_j] = result["PR(>F)"][1]
            data["direction p-values"]["p_values"][index_j, index_i] = result["PR(>F)"][1]

    ############ VISUALIZE ############

    visualize_sim_mat(
        data,
        mat_key="p_values",
        title="session_ANOVA_test",
        name_lst_key="measure_lst",
        annot=True,
        fmt=2,
        show_diag=False,
        show_sig=True,
        no_color=True,
        save_image=save_image,
        output_root=output_root + "dFC_similarity/",
    )

################# feature-based #################
"""
    - spatial
    - temporal
    - inter-time-correlation
    - inter-connection-correlation
    - dFC-avg
    - dFC-var
"""

for feature2extract in ALL_RESULTS["dFC_similarity_feature_based"]:

    RESULTS = ALL_RESULTS["dFC_similarity_feature_based"][feature2extract]
    filters_lst = [filter for filter in RESULTS]
    keys_lst = [key for key in RESULTS[filters_lst[0]]]

    ############ VISUALIZE ############
    label_dict = {
        "default_values": "session Rest1 LR",
        "session_Rest1_RL": "session Rest1 RL",
        "session_Rest2_LR": "session Rest2 LR",
        "session_Rest2_RL": "session Rest2 RL",
    }
    for key in keys_lst:
        if key == "name_lst":
            continue
        visualize_sim_mat(
            RESULTS,
            mat_key=key,
            title=feature2extract + " " + key,
            name_lst_key="name_lst",
            label_dict=label_dict,
            cmap="viridis",
            save_image=save_image,
            output_root=output_root + "feature_based/" + feature2extract + "/",
        )
    ############ Hierarchical Clustering ############
    for filter in ["default_values"]:
        dist_mat = corr2distance(RESULTS[filter]["avg_mat"], metric="spearman")
        Z = distance2Z(dist_mat, method="ward")

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(
            Z=Z,
            labels=RESULTS[filter]["name_lst"],
            cluster_colors_dict=cluster_colors_dict,
            extra_colors=extra_colors,
        )

        dist_mat_dendo(
            Z=Z,
            labels=RESULTS[filter]["name_lst"],
            link_colors=link_cols,
            plot_threshold=True,
            title="Hierarchical Clustering of Methods "
            + filter
            + " using "
            + feature2extract,
            save_image=save_image,
            output_root=output_root + "feature_based/" + feature2extract + "/",
        )

############ Spatial vs. Temporal Scatter plot ############

scatter_data = ALL_RESULTS["spatial_vs_temporal_similarity"]

############ visualization ############
label_dict = {"spatial": "spatial similarity", "temporal": "temporal similarity"}
scatter_plot(
    data=scatter_data,
    x="temporal",
    y="spatial",
    labels="labels",
    title="spatial similarity vs temporal similarity",
    label_dict=label_dict,
    c=0.3,
    equal_axis_lim=True,
    show_x_equal_y=True,
    save_image=save_image,
    output_root=output_root + "variation/spatialVsTemporal/",
)

################# graph-based #################
"""
    - spatial
    - temporal-avg
    - ECM, shortest_path, degree, clustering_coef
"""

###### spatial #####

for graph_property in ALL_RESULTS["dFC_similarity_graph"]["spatial"]:

    RESULTS = ALL_RESULTS["dFC_similarity_graph"]["spatial"][graph_property]
    filters_lst = [filter for filter in RESULTS]
    keys_lst = [key for key in RESULTS[filters_lst[0]]]

    ############ VISUALIZE ############
    label_dict = {
        "default_values": "session Rest1 LR",
        "session_Rest1_RL": "session Rest1 RL",
        "session_Rest2_LR": "session Rest2 LR",
        "session_Rest2_RL": "session Rest2 RL",
    }
    for key in keys_lst:
        if key == "name_lst":
            continue
        visualize_sim_mat(
            RESULTS,
            mat_key=key,
            title="spatial " + graph_property + " " + key,
            name_lst_key="name_lst",
            label_dict=label_dict,
            cmap="viridis",
            save_image=save_image,
            output_root=output_root + "graph_based/" + graph_property + "/",
        )
    ############ Hierarchical Clustering ############
    for filter in ["default_values"]:
        dist_mat = corr2distance(RESULTS[filter]["avg_mat"], metric="spearman")
        Z = distance2Z(dist_mat, method="ward")

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(
            Z=Z,
            labels=RESULTS[filter]["name_lst"],
            cluster_colors_dict=cluster_colors_dict,
            extra_colors=extra_colors,
        )

        dist_mat_dendo(
            Z=Z,
            labels=RESULTS[filter]["name_lst"],
            link_colors=link_cols,
            plot_threshold=True,
            title="Hierarchical Clustering of Methods "
            + filter
            + " using "
            + "spatial "
            + graph_property,
            save_image=save_image,
            output_root=output_root + "graph_based/" + graph_property + "/",
        )

################################# TSNE visualization #################################

"""
    - visualize subjects dFC obtained by different methods in a 2D space
        considering their distances
"""

RESULTS = ALL_RESULTS["TSNE"]
sample_measure_lst = [sample[sample.find("_") + 1 :] for sample in RESULTS["sample_lst"]]

measures_lst = list(set(sample_measure_lst))
measures_lst.sort()
colors_lst = ["black", "green", "orange", "red", "dodgerblue", "grey", "violet"]
color_dict = {}
for i, measure in enumerate(measures_lst):
    color_dict[measure] = colors_lst[i]

# using overall corr
dist_mat = corr2distance(RESULTS["corr"], metric="spearman")

plot_TSNE(
    dist_mat,
    sample_measure_lst,
    color_dict,
    projection="2d",
    title="TSNE overall corr",
    save_image=save_image,
    output_root=output_root + "TSNE/",
)

for n_components in RESULTS["X_red_corr"]:
    dist_mat = corr2distance(RESULTS["X_red_corr"][n_components], metric="spearman")

    plot_TSNE(
        dist_mat,
        sample_measure_lst,
        color_dict,
        projection="2d",
        title="TSNE " + str(n_components),
        save_image=save_image,
        output_root=output_root + "TSNE/",
    )

################################# inter_subject similarity #################################

"""
    - returns correspondence of inter-subject relation between results of dFC
        measures in each session
    - dendogram based on inter-subject similarity
"""

for subj_lvl_feature in ALL_RESULTS["subj_clustring"]:

    RESULTS = ALL_RESULTS["subj_clustring"][subj_lvl_feature]

    ############ VISUALIZE ############
    label_dict = {
        "session_Rest1_LR": "session Rest1 LR",
        "session_Rest1_RL": "session Rest1 RL",
        "session_Rest2_LR": "session Rest2 LR",
        "session_Rest2_RL": "session Rest2 RL",
    }
    for key in RESULTS:
        annot = True
        # set diag values to 0
        for session in RESULTS[key]:
            np.fill_diagonal(RESULTS[key][session]["sim_mat"], 0)
        visualize_sim_mat(
            RESULTS[key],
            mat_key="sim_mat",
            title="inter-subject-corr similarity "
            + key
            + " based on "
            + subj_lvl_feature,
            name_lst_key="name_lst",
            label_dict=label_dict,
            cmap="viridis",
            annot=annot,
            save_image=save_image,
            output_root=output_root + "inter_subject/" + subj_lvl_feature + "/",
        )
    ############ Hierarchical Clustering ############
    for session in RESULTS["across_method"]:
        dist_mat = corr2distance(
            RESULTS["across_method"][session]["sim_mat"], metric="spearman"
        )
        Z = distance2Z(dist_mat, method="ward")

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(
            Z=Z,
            labels=RESULTS["across_method"][session]["name_lst"],
            cluster_colors_dict=cluster_colors_dict,
            extra_colors=extra_colors,
        )

        dist_mat_dendo(
            Z=Z,
            labels=RESULTS["across_method"][session]["name_lst"],
            link_colors=link_cols,
            plot_threshold=True,
            title="Hierarchical Clustering of Methods "
            + session
            + " using inter-subject similarity based on "
            + subj_lvl_feature,
            save_image=save_image,
            output_root=output_root + "inter_subject/" + subj_lvl_feature + "/",
        )

################################# dFC var #################################

"""
    - avg(variance/fluctuations of dFC in one Subj)
    - rank normed
"""

RESULTS = ALL_RESULTS["dFC_var"]

visualize_conn_mat_dict(
    RESULTS["avg_dFC_var"],
    node_networks=node_networks,
    title="avg dFC var",
    center_0=False,
    fix_lim=False,
    disp_diag=True,
    cmap="plasma",
    normalize=False,
    save_image=save_image,
    output_root=output_root + "dFC_var/",
)

visualize_conn_mat_dict(
    RESULTS["avg_dFC_var"],
    node_networks=node_networks,
    segmented=True,
    title="segmented avg dFC var",
    center_0=False,
    fix_lim=False,
    disp_diag=True,
    cmap="plasma",
    normalize=False,
    save_image=save_image,
    output_root=output_root + "dFC_var/",
)

visualize_conn_mat_dict(
    RESULTS["var_dFC_var"],
    node_networks=node_networks,
    title="var of dFC var",
    center_0=False,
    fix_lim=False,
    disp_diag=True,
    cmap="plasma",
    normalize=False,
    save_image=save_image,
    output_root=output_root + "dFC_var/",
)

################################# dFC avg #################################

"""
    - avg(avg of dFC -static FC- in one Subj)
    - rank normed
"""

RESULTS = ALL_RESULTS["dFC_avg"]

visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    title="dFC avg",
    center_0=False,
    fix_lim=False,
    disp_diag=False,
    cmap="plasma",
    normalize=False,
    save_image=save_image,
    output_root=output_root + "dFC_avg/",
)

visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    segmented=True,
    title="segmented dFC avg",
    center_0=False,
    fix_lim=False,
    disp_diag=False,
    cmap="plasma",
    normalize=False,
    save_image=save_image,
    output_root=output_root + "dFC_avg/",
)

################################# Across Func Conn total Correlation #################################

"""
    - spearman_corr((dFConnection(node_i, node_j) timecourse using method m), (dFConnection(node_i, node_j) timecourse using method n))
"""

RESULTS = ALL_RESULTS["across_func_conns"]

############ VISUALIZE ############

visualize_conn_mat_2D_dict(
    RESULTS,
    node_networks=node_networks,
    title="across func conn total spearman corr",
    fix_lim=False,
    disp_diag=False,
    cmap="seismic",
    normalize=False,
    center_0=True,
    save_image=save_image,
    output_root=output_root + "across_func_conn/",
)

visualize_conn_mat_2D_dict(
    RESULTS,
    node_networks=node_networks,
    segmented=True,
    title="segmented across func conn total spearman corr",
    fix_lim=False,
    disp_diag=False,
    cmap="seismic",
    normalize=False,
    center_0=True,
    save_image=save_image,
    output_root=output_root + "across_func_conn/",
)

visualize_conn_mat_2D_dict(
    RESULTS,
    node_networks=node_networks,
    title="across func conn total spearman corr normalized",
    fix_lim=False,
    disp_diag=False,
    cmap="seismic",
    normalize=True,
    center_0=True,
    save_image=save_image,
    output_root=output_root + "across_func_conn/",
)

visualize_conn_mat_2D_dict(
    RESULTS,
    node_networks=node_networks,
    segmented=True,
    title="segmented across func conn total spearman corr normalized",
    fix_lim=False,
    disp_diag=False,
    cmap="seismic",
    normalize=True,
    center_0=True,
    save_image=save_image,
    output_root=output_root + "across_func_conn/",
)

################################# High Variation Regions #################################
"""
    - high variation regions over methods and over time.
"""

RESULTS = ALL_RESULTS["var_across_func_conns"]
RATIO = ALL_RESULTS["var_across_func_conns_ratio"]
scatter_data = ALL_RESULTS["var_method_vs_time_across_func_conns_scatter"]

############ method var / time var ############

ratio = list()
ratio_std = list()
for i, sample in enumerate(scatter_data["var_method"]):
    ratio.append(scatter_data["var_method"][i] / scatter_data["var_time"][i])
    ratio_std.append(
        np.sqrt(scatter_data["var_method"][i]) / np.sqrt(scatter_data["var_time"][i])
    )
ratio = np.mean(np.array(ratio))
ratio_std = np.mean(np.array(ratio_std))
avg_var_time = np.mean(scatter_data["var_time"])
avg_var_method = np.mean(scatter_data["var_method"])

# write to a txt file
folder = output_root + "variation/timeVsMethod"
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder + "/var_ratio.txt")
filename.touch(exist_ok=True)
text_file = open(filename, "wt")
text_file.write("Average of var method / var time ratio = " + str(ratio) + "\n")
text_file.write("Average of std method / std time ratio = " + str(ratio_std) + "\n")
text_file.write("Average var method = " + str(avg_var_method) + "\n")
text_file.write("Average var time = " + str(avg_var_time) + "\n")
text_file.write(
    "Ratio of avg var method / avg var time = "
    + str(avg_var_method / avg_var_time)
    + "\n"
)
text_file.close()

############ VISUALIZE ############

label_dict = {"var_method": "variation over method", "var_time": "variation over time"}
scatter_plot(
    data=scatter_data,
    x="var_time",
    y="var_method",
    title="var method vs time across func conns",
    label_dict=label_dict,
    hist=True,
    equal_axis_lim=True,
    show_x_equal_y=True,
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

label_dict = {
    "var across method / time - 1": "var over method / time - 1",
}
visualize_conn_mat_dict(
    RATIO,
    node_networks=node_networks,
    segmented=False,
    title="ratio of high variation regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=False,
    cmap="seismic",
    center_0=True,
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

visualize_conn_mat_dict(
    RATIO,
    node_networks=node_networks,
    segmented=True,
    title="segmented ratio of high variation regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=False,
    cmap="seismic",
    center_0=True,
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

label_dict = {
    "var_over_method": "variation over method",
    "var_over_time": "variation over time",
}

visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    title="variation across regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=True,
    cmap="plasma",
    center_0=False,
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

# func conn segmented
visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    segmented=True,
    title="segmented high variation regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=True,
    cmap="plasma",
    center_0=False,
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

RESULTS = ALL_RESULTS["high_var_func_conns"]

label_dict = {
    "var_over_method": "variation over method",
    "var_over_time": "variation over time",
}
visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    title="high variation regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=True,
    cmap="plasma",
    center_0=False,
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

################################# Variation Value Comparison #################################
"""
    - compare variation over methods with variation over time.
"""

RESULTS = ALL_RESULTS["var_comparison"]
scatter_data_across_func_conn = ALL_RESULTS[
    "var_method_vs_time_method_pairs_across_func_conns"
]
scatter_data = ALL_RESULTS["var_method_vs_time_method_pairs"]

############ VISUALIZE ############

visualize_sim_mat(
    RESULTS,
    mat_key="sim_mat",
    title="variation in different dimensions",
    name_lst_key="name_lst",
    cmap="viridis",
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

label_dict = {"var_method": "variation over methods", "var_time": "variation over time"}
pairwise_scatter_plots(
    data=scatter_data_across_func_conn,
    x="var_time",
    y="var_method",
    title="var method vs time across func conns across methods pairs",
    label_dict=label_dict,
    hist=True,
    equal_axis_lim=True,
    show_x_equal_y=True,
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

label_dict = {"var_method": "variation over method", "var_time": "variation over time"}
scatter_plot(
    data=scatter_data,
    x="var_time",
    y="var_method",
    labels="labels",
    title="var method vs time",
    label_dict=label_dict,
    c=0.5,
    equal_axis_lim=True,
    show_x_equal_y=True,
    save_image=save_image,
    output_root=output_root + "variation/timeVsMethod/",
)

################################# Var method vs. Time Clstrwise #################################
"""
    - compute var over method and time within each group of methods
"""
scatter_data = ALL_RESULTS["var_method_vs_time_clstrwise"]["scatter_data"]
clstrs_dict = ALL_RESULTS["var_method_vs_time_clstrwise"]["clstrs_dict"]

# write to a txt file
folder = output_root + "variation"
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder + "/var_clstrwise_ratio.txt")
filename.touch(exist_ok=True)
text_file = open(filename, "wt")

for clstr in scatter_data:

    ratio = list()
    for i, sample in enumerate(scatter_data[clstr]["var_method"]):
        ratio.append(
            scatter_data[clstr]["var_method"][i] / scatter_data[clstr]["var_time"][i]
        )

    ratio = np.mean(np.array(ratio))
    avg_var_time = np.mean(scatter_data[clstr]["var_time"])
    avg_var_method = np.mean(scatter_data[clstr]["var_method"])

    text_file.write(clstr + "\n")
    text_file.write(" ".join(clstrs_dict[clstr]) + "\n")
    text_file.write("Average of var method / var time ratio = " + str(ratio) + "\n")
    text_file.write("Average var method = " + str(avg_var_method) + "\n")
    text_file.write("Average var time = " + str(avg_var_time) + "\n")
    text_file.write(
        "Ratio of avg var method / avg var time = "
        + str(avg_var_method / avg_var_time)
        + "\n"
    )

    label_dict = {
        "var_method": "variation over method",
        "var_time": "variation over time",
    }
    scatter_plot(
        data=scatter_data[clstr],
        x="var_time",
        y="var_method",
        title="var method vs time clusterwise across func conns " + clstr,
        label_dict=label_dict,
        hist=True,
        equal_axis_lim=True,
        show_x_equal_y=True,
        save_image=save_image,
        output_root=output_root + "variation/timeVsMethod/",
    )
text_file.close()

################################# Inter-subject Variation #################################
"""
    - compute var over method and subject across all func conns
"""

RESULTS = ALL_RESULTS["inter-subj_var_across_func_conns"]
RATIO = ALL_RESULTS["inter-subj_var_across_func_conns_ratio"]
scatter_data = ALL_RESULTS["var_method_vs_subj_across_func_conns_scatter"]

############ method var / subject var ############

ratio = list()
ratio_std = list()
for i, sample in enumerate(scatter_data["var_method"]):
    ratio.append(scatter_data["var_method"][i] / scatter_data["var_subj"][i])
    ratio_std.append(
        np.sqrt(scatter_data["var_method"][i]) / np.sqrt(scatter_data["var_subj"][i])
    )
ratio = np.mean(np.array(ratio))
ratio_std = np.mean(np.array(ratio_std))
avg_var_subj = np.mean(scatter_data["var_subj"])
avg_var_method = np.mean(scatter_data["var_method"])

# write to a txt file
folder = output_root + "variation/subjVsMethod"
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder + "/subj_var_ratio.txt")
filename.touch(exist_ok=True)
text_file = open(filename, "wt")
text_file.write("Average of var method / var subj ratio = " + str(ratio) + "\n")
text_file.write("Average of std method / std subj ratio = " + str(ratio_std) + "\n")
text_file.write("Average var method = " + str(avg_var_method) + "\n")
text_file.write("Average var subj = " + str(avg_var_subj) + "\n")
text_file.write(
    "Ratio of avg var method / avg var subj = "
    + str(avg_var_method / avg_var_subj)
    + "\n"
)
text_file.close()

############ VISUALIZE ############

label_dict = {"var_method": "variation over method", "var_subj": "variation over subject"}
scatter_plot(
    data=scatter_data,
    x="var_subj",
    y="var_method",
    title="var method vs subj across func conns",
    label_dict=label_dict,
    hist=True,
    equal_axis_lim=True,
    show_x_equal_y=True,
    save_image=save_image,
    output_root=output_root + "variation/subjVsMethod/",
)

label_dict = {
    "var across method / subj - 1": "var over method / subj - 1",
}
visualize_conn_mat_dict(
    RATIO,
    node_networks=node_networks,
    segmented=False,
    title="ratio of method subj var regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=False,
    cmap="seismic",
    center_0=True,
    save_image=save_image,
    output_root=output_root + "variation/subjVsMethod/",
)

visualize_conn_mat_dict(
    RATIO,
    node_networks=node_networks,
    segmented=True,
    title="segmented ratio of method subj var regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=False,
    cmap="seismic",
    center_0=True,
    save_image=save_image,
    output_root=output_root + "variation/subjVsMethod/",
)

label_dict = {
    "var_over_method": "variation over method",
    "var_over_subj": "variation over subject",
}

visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    title="inter-subj variation across regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=True,
    cmap="plasma",
    center_0=False,
    save_image=save_image,
    output_root=output_root + "variation/subjVsMethod/",
)

# func conn segmented
visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    segmented=True,
    title="segmented inter-subj variation across regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=True,
    cmap="plasma",
    center_0=False,
    save_image=save_image,
    output_root=output_root + "variation/subjVsMethod/",
)

################################# Overall Variability #################################
"""
    - compute var over method and overall across all func conns
"""

RESULTS = ALL_RESULTS["var_all_across_func_conns"]
RATIO = ALL_RESULTS["var_all_across_func_conns_ratio"]
scatter_data = ALL_RESULTS["var_method_vs_all_across_func_conns_scatter"]

############ method var / subject var ############

ratio = list()
ratio_std = list()
for i, sample in enumerate(scatter_data["var_method"]):
    ratio.append(scatter_data["var_method"][i] / scatter_data["var_all"][i])
    ratio_std.append(
        np.sqrt(scatter_data["var_method"][i]) / np.sqrt(scatter_data["var_all"][i])
    )
ratio = np.mean(np.array(ratio))
ratio_std = np.mean(np.array(ratio_std))
avg_var_all = np.mean(scatter_data["var_all"])
avg_var_method = np.mean(scatter_data["var_method"])

# write to a txt file
folder = output_root + "variation/allVsMethod"
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder + "/all_var_ratio.txt")
filename.touch(exist_ok=True)
text_file = open(filename, "wt")
text_file.write("Average of var method / var all ratio = " + str(ratio) + "\n")
text_file.write("Average of std method / std all ratio = " + str(ratio_std) + "\n")
text_file.write("Average var method = " + str(avg_var_method) + "\n")
text_file.write("Average var subj = " + str(avg_var_all) + "\n")
text_file.write(
    "Ratio of avg var method / avg var subj = " + str(avg_var_method / avg_var_all) + "\n"
)
text_file.close()

############ VISUALIZE ############

label_dict = {"var_method": "variation over method", "var_all": "variation over all"}
scatter_plot(
    data=scatter_data,
    x="var_all",
    y="var_method",
    title="var method vs all across func conns",
    label_dict=label_dict,
    hist=True,
    equal_axis_lim=True,
    show_x_equal_y=True,
    save_image=save_image,
    output_root=output_root + "variation/allVsMethod/",
)

label_dict = {
    "var across method / all - 1": "var over method / all - 1",
}
visualize_conn_mat_dict(
    RATIO,
    node_networks=node_networks,
    segmented=False,
    title="ratio of method all var regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=False,
    cmap="seismic",
    center_0=True,
    save_image=save_image,
    output_root=output_root + "variation/allVsMethod/",
)

visualize_conn_mat_dict(
    RATIO,
    node_networks=node_networks,
    segmented=True,
    title="segmented ratio of method all var regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=False,
    cmap="seismic",
    center_0=True,
    save_image=save_image,
    output_root=output_root + "variation/allVsMethod/",
)

label_dict = {
    "var_over_method": "variation over method",
    "var_over_all": "variation over all",
}

visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    title="all variation across regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=True,
    cmap="plasma",
    center_0=False,
    save_image=save_image,
    output_root=output_root + "variation/allVsMethod/",
)

# func conn segmented
visualize_conn_mat_dict(
    RESULTS,
    node_networks=node_networks,
    segmented=True,
    title="segmented all variation across regions",
    fix_lim=False,
    label_dict=label_dict,
    disp_diag=True,
    cmap="plasma",
    center_0=False,
    save_image=save_image,
    output_root=output_root + "variation/allVsMethod/",
)

################################# Randomization Tests #################################

metric = "spearman"

"""
find the similarity between the dFC obtained by each method
and a dFC which is a constant sequence of mean of all methods dFC
"""

########### Similarity with static FC ###########

RESULTS = ALL_RESULTS["randomization"]["sim_with_static_FC"]

############ VISUALIZE ############
label_dict = {"sim": "similarity", "dFC_method": "dFC assessment method"}
cat_plot(
    data=RESULTS,
    x="dFC_method",
    y="sim",
    kind="box",
    title="similarity with constant static FC",
    label_dict=label_dict,
    save_image=save_image,
    output_root=output_root + "randomization/",
)

########### Shuffled ###########

"""
find the similarity between the dFC obtained by each method
but with randomized temporal/spatial/all order
"""

RESULTS = ALL_RESULTS["randomization"]["shuffled"]["similarity_dict"]
modes_lst = ALL_RESULTS["randomization"]["shuffled"]["modes_lst"]

############ VISUALIZE ############
label_dict = {
    "all_shuffle_sim": "similarity",
    "spatial_shuffle_sim": "similarity",
    "temporal_shuffle_sim": "similarity",
}
for mode in modes_lst:
    key = mode + "_shuffle_sim"
    pairwise_cat_plots(
        RESULTS,
        y=key,
        z="actual_sim",
        title=mode + " shuffled similarity",
        label_dict=label_dict,
        save_image=save_image,
        output_root=output_root + "randomization/",
    )

########### Random state time course ###########

"""
find the similarity between dFC matrices created using
a random state time course but real FC patterns obtained
by each method; for SW and TF all FC patterns of each
subject are used
"""

RESULTS = ALL_RESULTS["randomization"]["random_state_TC"]

############ VISUALIZE ############
label_dict = {
    "sim": "similarity",
}
pairwise_cat_plots(
    RESULTS,
    y="sim",
    z="actual_sim",
    title="random state time course dFC",
    label_dict=label_dict,
    save_image=save_image,
    output_root=output_root + "randomization/",
)

################################# inter-state spatial similarity #################################

key_name = "inter-state_spatial_sim"
RESULTS = ALL_RESULTS[key_name]

############ VISUALIZE ############
label_dict = {
    "sim": "inter-state spatial similarity",
    "dFC_method": "dFC assessment method",
}
cat_plot(
    data=RESULTS,
    x="dFC_method",
    y="sim",
    kind="box",
    scale_dist=False,
    title=key_name,
    label_dict=label_dict,
    save_image=save_image,
    output_root=output_root + "indiv_prop/",
)

################################# SIMILARITY OF ADJACENT TIME POINTS #################################

key_name = "similarity of adjacent time points"
RESULTS = ALL_RESULTS["adjacent_time_points"]

############ VISUALIZE ############
label_dict = {"dFC_method": "dFC assessment method"}
cat_plot(
    data=RESULTS,
    x="dFC_method",
    y=key_name,
    kind="box",
    title=key_name,
    label_dict=label_dict,
    save_image=save_image,
    output_root=output_root + "indiv_prop/",
)

################################# TRANSITION FREQUENCY #################################
"""
 - plot normalized transition frequency
"""

key_name = "trans_freq"
RESULTS = ALL_RESULTS["transition_freq"]

############ VISUALIZE ############
label_dict = {"trans_freq": "transition frequency", "dFC_method": "dFC assessment method"}
cat_plot(
    data=RESULTS,
    x="dFC_method",
    y=key_name,
    kind="box",
    title=key_name,
    label_dict=label_dict,
    save_image=save_image,
    output_root=output_root + "indiv_prop/",
)

################################# DWELL TIME #################################
"""
 - plot normalized (not downsampled) dwell times
 - the dwell times are not averaged within each subj, they include all individual
 dwells
"""

key_name = "dwell_time"
RESULTS = ALL_RESULTS["dwell_time"]

############ VISUALIZE ############
label_dict = {"dwell_time": "dwell time", "dFC_method": "dFC assessment method"}
cat_plot(
    data=RESULTS,
    x="dFC_method",
    y=key_name,
    kind="box",
    scale_dist=True,
    title=key_name,
    label_dict=label_dict,
    y_lim=(0, 0.2),
    save_image=save_image,
    output_root=output_root + "indiv_prop/",
)

################################# TIME RECORD #################################

RESULTS = ALL_RESULTS["time_record"]
filter = "session_Rest1_LR"

############ VISUALIZE ############
label_dict = {
    "FCS_fit_time (s)": "FCS fitting time (s)",
    "dFC_assess_time (s)": "dFC assessing time (s)",
    "dFC_method": "dFC assessment method",
}
cat_plot(
    data=RESULTS,
    x="dFC_method",
    y="dFC_assess_time (s)",
    kind="bar",
    log=True,
    title="dFC assess time record of " + filter,
    label_dict=label_dict,
    save_image=save_image,
    output_root=output_root + "time/",
)

cat_plot(
    data=RESULTS,
    x="dFC_method",
    y="FCS_fit_time (s)",
    kind="bar",
    log=True,
    title="FCS fit time record of " + filter,
    label_dict=label_dict,
    save_image=save_image,
    output_root=output_root + "time/",
)

if not save_image:
    plt.show()
#################################################################################
